# Copyright 2019 NetApp, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

import ddt
from oslo_db import exception as db_exception

from manila.api import common
from manila.api.v2 import share_network_subnets
from manila.db import api as db_api
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils
from webob import exc

fake_az = {
    'id': 'ae525e12-07e8-4ddc-a2fd-4a89ad4a65ff',
    'name': 'fake_az_name'
}

fake_default_subnet = {
    'neutron_net_id': 'fake_nn_id',
    'neutron_subnet_id': 'fake_nsn_id',
    'availability_zone_id': None
}

fake_subnet_with_az = {
    'neutron_net_id': 'fake_nn_id',
    'neutron_subnet_id': 'fake_nsn_id',
    'availability_zone_id': 'fake_az_id'
}


@ddt.ddt
class ShareNetworkSubnetControllerTest(test.TestCase):
    """Share network subnet api test"""

    def setUp(self):
        super(ShareNetworkSubnetControllerTest, self).setUp()
        self.controller = share_network_subnets.ShareNetworkSubnetController()
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.resource_name = self.controller.resource_name
        self.mock_az_get = self.mock_object(db_api, 'availability_zone_get',
                                            mock.Mock(return_value=fake_az))
        self.share_network = db_utils.create_share_network(
            name='fake_network', id='fake_sn_id')
        self.share_server = db_utils.create_share_server(
            share_network_subnet_id='fake_sns_id')
        self.subnet = db_utils.create_share_network_subnet(
            share_network_id=self.share_network['id'])
        self.share = db_utils.create_share()

    def test_share_network_subnet_delete(self):
        req = fakes.HTTPRequest.blank('/subnets/%s' % self.subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']

        mock_sns_get = self.mock_object(
            db_api, 'share_network_subnet_get',
            mock.Mock(return_value=self.subnet))
        mock_all_get_all_shares_by_ss = self.mock_object(
            db_api, 'share_instances_get_all_by_share_server',
            mock.Mock(return_value=[]))
        mock_all_ss_are_auto_deletable = self.mock_object(
            self.controller, '_all_share_servers_are_auto_deletable',
            mock.Mock(return_value=True))
        mock_delete_share_server = self.mock_object(
            self.controller.share_rpcapi, 'delete_share_server')
        mock_subnet_delete = self.mock_object(db_api,
                                              'share_network_subnet_delete')

        result = self.controller.delete(req, self.share_network['id'],
                                        self.subnet['id'])

        self.assertEqual(202, result.status_int)
        mock_sns_get.assert_called_once_with(
            context, self.subnet['id'])
        mock_all_get_all_shares_by_ss.assert_called_once_with(
            context, self.subnet['share_servers'][0].id
        )
        mock_all_ss_are_auto_deletable.assert_called_once_with(
            self.subnet)
        mock_delete_share_server.assert_called_once_with(
            context, self.subnet['share_servers'][0])
        mock_subnet_delete.assert_called_once_with(
            context, self.subnet['id'])
        policy.check_policy.assert_called_once_with(
            context, self.resource_name, 'delete')

    def test_share_network_subnet_delete_network_not_found(self):
        req = fakes.HTTPRequest.blank('/subnets/%s' % self.subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']

        mock_sn_get = self.mock_object(
            db_api, 'share_network_get',
            mock.Mock(side_effect=exception.ShareNetworkNotFound(
                share_network_id=self.share_network['id']
            )))

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.delete,
                          req,
                          self.share_network['id'],
                          self.subnet['id'])
        mock_sn_get.assert_called_once_with(
            context, self.share_network['id'])
        self.mock_policy_check.assert_called_once_with(
            context, self.resource_name, 'delete')

    def test_share_network_subnet_delete_subnet_not_found(self):
        req = fakes.HTTPRequest.blank('/subnets/%s' % self.subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']

        mock_sns_get = self.mock_object(
            db_api, 'share_network_subnet_get',
            mock.Mock(side_effect=exception.ShareNetworkSubnetNotFound(
                share_network_subnet_id=self.subnet['id']
            )))

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.delete,
                          req,
                          self.share_network['id'],
                          self.subnet['id'])
        mock_sns_get.assert_called_once_with(
            context, self.subnet['id'])
        self.mock_policy_check.assert_called_once_with(
            context, self.resource_name, 'delete')

    def test_delete_subnet_with_share_servers_fail(self):
        req = fakes.HTTPRequest.blank('/subnets/%s' % self.subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']

        mock_sns_get = self.mock_object(
            db_api, 'share_network_subnet_get',
            mock.Mock(return_value=self.subnet))
        mock_all_get_all_shares_by_ss = self.mock_object(
            db_api, 'share_instances_get_all_by_share_server',
            mock.Mock(return_value=[]))
        mock_all_ss_are_auto_deletable = self.mock_object(
            self.controller, '_all_share_servers_are_auto_deletable',
            mock.Mock(return_value=False))

        self.assertRaises(exc.HTTPConflict,
                          self.controller.delete,
                          req,
                          self.share_network['id'],
                          self.subnet['id'])

        mock_sns_get.assert_called_once_with(
            context, self.subnet['id'])
        mock_all_get_all_shares_by_ss.assert_called_once_with(
            context, self.subnet['share_servers'][0].id
        )
        mock_all_ss_are_auto_deletable.assert_called_once_with(
            self.subnet
        )
        self.mock_policy_check.assert_called_once_with(
            context, self.resource_name, 'delete')

    def test_delete_subnet_with_shares_fail(self):
        req = fakes.HTTPRequest.blank('/subnets/%s' % self.subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']

        mock_sns_get = self.mock_object(
            db_api, 'share_network_subnet_get',
            mock.Mock(return_value=self.subnet))
        mock_all_get_all_shares_by_ss = self.mock_object(
            db_api, 'share_instances_get_all_by_share_server',
            mock.Mock(return_value=[self.share]))

        self.assertRaises(exc.HTTPConflict,
                          self.controller.delete,
                          req,
                          self.share_network['id'],
                          self.subnet['id'])

        mock_sns_get.assert_called_once_with(
            context, self.subnet['id'])
        mock_all_get_all_shares_by_ss.assert_called_once_with(
            context, self.subnet['share_servers'][0].id
        )
        self.mock_policy_check.assert_called_once_with(
            context, self.resource_name, 'delete')

    @ddt.data((None, fake_default_subnet, None),
              (fake_az, None, fake_subnet_with_az))
    @ddt.unpack
    def test__validate_subnet(self, az, default_subnet, subnet_az):
        req = fakes.HTTPRequest.blank('/subnets', version='2.51')
        context = req.environ['manila.context']

        mock_get_default_sns = self.mock_object(
            db_api, 'share_network_subnet_get_default_subnet',
            mock.Mock(return_value=default_subnet))
        mock_get_subnet_by_az = self.mock_object(
            db_api, 'share_network_subnet_get_by_availability_zone_id',
            mock.Mock(return_value=subnet_az))

        self.assertRaises(exc.HTTPConflict,
                          self.controller._validate_subnet,
                          context,
                          self.share_network['id'],
                          az)
        if az:
            mock_get_subnet_by_az.assert_called_once_with(
                context, self.share_network['id'], az['id'])
            mock_get_default_sns.assert_not_called()
        else:
            mock_get_default_sns.assert_called_once_with(
                context, self.share_network['id'])
            mock_get_subnet_by_az.assert_not_called()

    def _setup_create_test_request_body(self):
        body = {
            'share_network_id': self.share_network['id'],
            'availability_zone': fake_az['name'],
            'neutron_net_id': 'fake_nn_id',
            'neutron_subnet_id': 'fake_nsn_id'
        }
        return body

    def test_subnet_create(self):
        req = fakes.HTTPRequest.blank('/subnets', version="2.51")
        context = req.environ['manila.context']
        body = {
            'share-network-subnet': self._setup_create_test_request_body()
        }
        sn_id = body['share-network-subnet']['share_network_id']

        expected_result = copy.deepcopy(body)
        expected_result['share-network-subnet']['id'] = self.subnet['id']
        mock_check_net_and_subnet_id = self.mock_object(
            common, 'check_net_id_and_subnet_id')
        mock_validate_subnet = self.mock_object(
            self.controller, '_validate_subnet')
        mock_subnet_create = self.mock_object(
            db_api, 'share_network_subnet_create',
            mock.Mock(return_value=self.subnet))

        self.controller.create(
            req, body['share-network-subnet']['share_network_id'], body)

        mock_check_net_and_subnet_id.assert_called_once_with(
            body['share-network-subnet'])
        mock_validate_subnet.assert_called_once_with(
            context, sn_id, az=fake_az)
        mock_subnet_create.assert_called_once_with(
            context, body['share-network-subnet'])

    def test_subnet_create_share_network_not_found(self):
        fake_sn_id = 'fake_id'
        req = fakes.HTTPRequest.blank('/subnets', version="2.51")
        context = req.environ['manila.context']
        body = {
            'share-network-subnet': self._setup_create_test_request_body()
        }
        mock_sn_get = self.mock_object(
            db_api, 'share_network_get',
            mock.Mock(side_effect=exception.ShareNetworkNotFound(
                share_network_id=fake_sn_id)))

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.create,
                          req,
                          fake_sn_id,
                          body)
        mock_sn_get.assert_called_once_with(context, fake_sn_id)

    def test_subnet_create_az_not_found(self):
        fake_sn_id = 'fake_id'
        req = fakes.HTTPRequest.blank('/subnets', version="2.51")
        context = req.environ['manila.context']
        body = {
            'share-network-subnet': self._setup_create_test_request_body()
        }
        mock_sn_get = self.mock_object(db_api, 'share_network_get')
        mock_az_get = self.mock_object(
            db_api, 'availability_zone_get',
            mock.Mock(side_effect=exception.AvailabilityZoneNotFound(id='')))

        expected_az = body['share-network-subnet']['availability_zone']

        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          req,
                          fake_sn_id,
                          body)
        mock_sn_get.assert_called_once_with(context, fake_sn_id)
        mock_az_get.assert_called_once_with(
            context, expected_az)

    def test_subnet_create_subnet_default_or_same_az_exists(self):
        fake_sn_id = 'fake_id'
        req = fakes.HTTPRequest.blank('/subnets', version="2.51")
        context = req.environ['manila.context']
        body = {
            'share-network-subnet': self._setup_create_test_request_body()
        }
        mock_sn_get = self.mock_object(db_api, 'share_network_get')
        mock__validate_subnet = self.mock_object(
            self.controller, '_validate_subnet',
            mock.Mock(side_effect=exc.HTTPConflict(''))
        )
        expected_az = body['share-network-subnet']['availability_zone']

        self.assertRaises(exc.HTTPConflict,
                          self.controller.create,
                          req,
                          fake_sn_id,
                          body)
        mock_sn_get.assert_called_once_with(context, fake_sn_id)
        self.mock_az_get.assert_called_once_with(context, expected_az)
        mock__validate_subnet.assert_called_once_with(
            context, fake_sn_id, az=fake_az)

    def test_subnet_create_subnet_db_error(self):
        fake_sn_id = 'fake_sn_id'
        req = fakes.HTTPRequest.blank('/subnets', version="2.51")
        context = req.environ['manila.context']
        body = {
            'share-network-subnet': self._setup_create_test_request_body()
        }
        mock_sn_get = self.mock_object(db_api, 'share_network_get')
        mock__validate_subnet = self.mock_object(
            self.controller, '_validate_subnet')
        mock_db_subnet_create = self.mock_object(
            db_api, 'share_network_subnet_create',
            mock.Mock(side_effect=db_exception.DBError()))
        expected_data = copy.deepcopy(body['share-network-subnet'])
        expected_data['availability_zone_id'] = fake_az['id']
        expected_data.pop('availability_zone')

        self.assertRaises(exc.HTTPInternalServerError,
                          self.controller.create,
                          req,
                          fake_sn_id,
                          body)

        mock_sn_get.assert_called_once_with(context, fake_sn_id)
        self.mock_az_get.assert_called_once_with(context, fake_az['name'])
        mock__validate_subnet.assert_called_once_with(
            context, fake_sn_id, az=fake_az)
        mock_db_subnet_create.assert_called_once_with(
            context, expected_data
        )

    def test_show_subnet(self):
        subnet = db_utils.create_share_network_subnet(
            id='fake_sns_2', share_network_id=self.share_network['id'])
        expected_result = {
            'share_network_subnet': {
                "created_at": subnet['created_at'],
                "id": subnet['id'],
                "share_network_id": subnet['share_network_id'],
                "share_network_name": self.share_network['name'],
                "availability_zone": subnet['availability_zone'],
                "segmentation_id": subnet['segmentation_id'],
                "neutron_subnet_id": subnet['neutron_subnet_id'],
                "updated_at": subnet['updated_at'],
                "neutron_net_id": subnet['neutron_net_id'],
                "ip_version": subnet['ip_version'],
                "cidr": subnet['cidr'],
                "network_type": subnet['network_type'],
                "gateway": subnet['gateway'],
                "mtu": subnet['mtu'],
            }
        }
        req = fakes.HTTPRequest.blank('/subnets/%s' % subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']
        mock_sn_get = self.mock_object(
            db_api, 'share_network_get', mock.Mock(
                return_value=self.share_network))
        mock_sns_get = self.mock_object(
            db_api, 'share_network_subnet_get', mock.Mock(
                return_value=subnet))

        result = self.controller.show(req, self.share_network['id'],
                                      subnet['id'])

        self.assertEqual(expected_result, result)
        mock_sn_get.assert_called_once_with(context, self.share_network['id'])
        mock_sns_get.assert_called_once_with(context, subnet['id'])

    @ddt.data(
        (mock.Mock(side_effect=exception.ShareNetworkNotFound(
            share_network_id='fake_net_id')), None),
        (mock.Mock(), mock.Mock(
            side_effect=exception.ShareNetworkSubnetNotFound(
                share_network_subnet_id='fake_subnet_id'))))
    @ddt.unpack
    def test_show_subnet_not_found(self, sn_get_side_effect,
                                   sns_get_side_effect):
        req = fakes.HTTPRequest.blank('/subnets/%s' % self.subnet['id'],
                                      version="2.51")
        context = req.environ['manila.context']

        mock_sn_get = self.mock_object(
            db_api, 'share_network_get', sn_get_side_effect)
        mock_sns_get = self.mock_object(
            db_api, 'share_network_subnet_get', sns_get_side_effect)

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show,
                          req,
                          self.share_network['id'],
                          self.subnet['id'])
        mock_sn_get.assert_called_once_with(context, self.share_network['id'])
        if sns_get_side_effect:
            mock_sns_get.assert_called_once_with(context, self.subnet['id'])

    def test_list_subnet(self):
        share_network_id = 'fake_id'
        subnet = db_utils.create_share_network_subnet(
            share_network_id=share_network_id, id='fake_id')
        fake_sn = db_utils.create_share_network(id=share_network_id)
        expected_result = {
            'share_network_subnets': [{
                "created_at": subnet['created_at'],
                "id": subnet['id'],
                "share_network_id": subnet['id'],
                "share_network_name": fake_sn["name"],
                "availability_zone": subnet['availability_zone'],
                "segmentation_id": subnet['segmentation_id'],
                "neutron_subnet_id": subnet['neutron_subnet_id'],
                "updated_at": subnet['updated_at'],
                "neutron_net_id": subnet['neutron_net_id'],
                "ip_version": subnet['ip_version'],
                "cidr": subnet['cidr'],
                "network_type": subnet['network_type'],
                "gateway": subnet['gateway'],
                "mtu": subnet['mtu'],
            }]
        }

        req = fakes.HTTPRequest.blank('/subnets/', version="2.51")
        context = req.environ['manila.context']
        mock_sn_get = self.mock_object(
            db_api, 'share_network_get', mock.Mock(
                return_value=fake_sn))

        result = self.controller.index(req, self.share_network['id'])

        self.assertEqual(expected_result, result)
        mock_sn_get.assert_called_once_with(context, self.share_network['id'])

    def test_list_subnet_share_network_not_found(self):
        req = fakes.HTTPRequest.blank('/subnets/', version="2.51")
        context = req.environ['manila.context']

        mock_sn_get = self.mock_object(
            db_api, 'share_network_get', mock.Mock(
                side_effect=exception.ShareNetworkNotFound(
                    share_network_id=self.share_network['id'])))

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.index,
                          req,
                          self.share_network['id'])
        mock_sn_get.assert_called_once_with(context, self.share_network['id'])



# Copyright 2011 OpenStack LLC.
# Copyright (c) 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
from oslo_utils import strutils
from six.moves import http_client
from six.moves.urllib import parse
import webob

from manila.api.openstack import api_version_request as api_version
from manila.api.openstack import wsgi
from manila.api.views import quota_sets as quota_sets_views
from manila import db
from manila import exception
from manila.i18n import _
from manila import quota

QUOTAS = quota.QUOTAS
LOG = log.getLogger(__name__)
NON_QUOTA_KEYS = ('tenant_id', 'id', 'force', 'share_type')


class QuotaSetsMixin(object):
    """The Quota Sets API controller common logic.

    Mixin class that should be inherited by Quota Sets API controllers,
    which are used for different API URLs and microversions.
    """

    resource_name = "quota_set"
    _view_builder_class = quota_sets_views.ViewBuilder

    @staticmethod
    def _validate_quota_limit(limit, minimum, maximum, force_update):
        # NOTE: -1 is a flag value for unlimited
        if limit < -1:
            msg = _("Quota limit must be -1 or greater.")
            raise webob.exc.HTTPBadRequest(explanation=msg)
        if ((limit < minimum and not force_update) and
           (maximum != -1 or (maximum == -1 and limit != -1))):
            msg = _("Quota limit must be greater than %s.") % minimum
            raise webob.exc.HTTPBadRequest(explanation=msg)
        if maximum != -1 and limit > maximum and not force_update:
            msg = _("Quota limit must be less than %s.") % maximum
            raise webob.exc.HTTPBadRequest(explanation=msg)

    @staticmethod
    def _validate_user_id_and_share_type_args(user_id, share_type):
        if user_id and share_type:
            msg = _("'user_id' and 'share_type' values are mutually exclusive")
            raise webob.exc.HTTPBadRequest(explanation=msg)

    @staticmethod
    def _get_share_type_id(context, share_type_name_or_id):
        if share_type_name_or_id:
            share_type = db.share_type_get_by_name_or_id(
                context, share_type_name_or_id)
            if share_type:
                return share_type['id']
            msg = _("Share type with name or id '%s' not found.") % (
                share_type_name_or_id)
            raise webob.exc.HTTPNotFound(explanation=msg)

    @staticmethod
    def _ensure_share_type_arg_is_absent(req):
        params = parse.parse_qs(req.environ.get('QUERY_STRING', ''))
        share_type = params.get('share_type', [None])[0]
        if share_type:
            msg = _("'share_type' key is not supported by this microversion. "
                    "Use 2.39 or greater microversion to be able "
                    "to use 'share_type' quotas.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

    @staticmethod
    def _ensure_specific_microversion_args_are_absent(body, keys,
                                                      microversion):
        body = body.get('quota_set', body)
        for key in keys:
            if body.get(key):
                msg = (_("'%(key)s' key is not supported by this "
                         "microversion. Use %(microversion)s or greater "
                         "microversion to be able to use '%(key)s' quotas.") %
                       {"key": key, "microversion": microversion})
                raise webob.exc.HTTPBadRequest(explanation=msg)

    def _get_quotas(self, context, project_id, user_id=None,
                    share_type_id=None, usages=False):
        self._validate_user_id_and_share_type_args(user_id, share_type_id)
        if user_id:
            values = QUOTAS.get_user_quotas(
                context, project_id, user_id, usages=usages)
        elif share_type_id:
            values = QUOTAS.get_share_type_quotas(
                context, project_id, share_type_id, usages=usages)
        else:
            values = QUOTAS.get_project_quotas(
                context, project_id, usages=usages)
        if usages:
            return values
        return {k: v['limit'] for k, v in values.items()}

    @wsgi.Controller.authorize("show")
    def _show(self, req, id, detail=False):
        context = req.environ['manila.context']
        params = parse.parse_qs(req.environ.get('QUERY_STRING', ''))
        user_id = params.get('user_id', [None])[0]
        share_type = params.get('share_type', [None])[0]
        try:
            db.authorize_project_context(context, id)
            # _get_quotas use 'usages' to indicate whether retrieve additional
            # attributes, so pass detail to the argument.
            share_type_id = self._get_share_type_id(context, share_type)
            quotas = self._get_quotas(
                context, id, user_id, share_type_id, usages=detail)
            return self._view_builder.detail_list(
                req, quotas, id, share_type_id)
        except exception.NotAuthorized:
            raise webob.exc.HTTPForbidden()

    @wsgi.Controller.authorize('show')
    def _defaults(self, req, id):
        context = req.environ['manila.context']
        return self._view_builder.detail_list(
            req, QUOTAS.get_defaults(context), id)

    @wsgi.Controller.authorize("update")
    def _update(self, req, id, body):
        context = req.environ['manila.context']
        project_id = id
        bad_keys = []
        force_update = False
        params = parse.parse_qs(req.environ.get('QUERY_STRING', ''))
        user_id = params.get('user_id', [None])[0]
        share_type = params.get('share_type', [None])[0]
        self._validate_user_id_and_share_type_args(user_id, share_type)
        share_type_id = self._get_share_type_id(context, share_type)
        body = body.get('quota_set', {})
        if share_type and body.get('share_groups',
                                   body.get('share_group_snapshots')):
            msg = _("Share type quotas cannot constrain share groups and "
                    "share group snapshots.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        try:
            settable_quotas = QUOTAS.get_settable_quotas(
                context, project_id, user_id=user_id,
                share_type_id=share_type_id)
        except exception.NotAuthorized:
            raise webob.exc.HTTPForbidden()

        for key, value in body.items():
            if key == 'share_networks' and share_type_id:
                msg = _("'share_networks' quota cannot be set for share type. "
                        "It can be set only for project or user.")
                raise webob.exc.HTTPBadRequest(explanation=msg)
            elif (key not in QUOTAS and key not in NON_QUOTA_KEYS):
                bad_keys.append(key)
            elif key == 'force':
                force_update = strutils.bool_from_string(value)
            elif key not in NON_QUOTA_KEYS and value:
                try:
                    value = int(value)
                except (ValueError, TypeError):
                    msg = _("Quota '%(value)s' for %(key)s should be "
                            "integer.") % {'value': value, 'key': key}
                    LOG.warning(msg)
                    raise webob.exc.HTTPBadRequest(explanation=msg)

        LOG.debug("Force update quotas: %s.", force_update)

        if len(bad_keys) > 0:
            msg = _("Bad key(s) %s in quota_set.") % ",".join(bad_keys)
            raise webob.exc.HTTPBadRequest(explanation=msg)

        try:
            quotas = self._get_quotas(
                context, id, user_id=user_id, share_type_id=share_type_id,
                usages=True)
        except exception.NotAuthorized:
            raise webob.exc.HTTPForbidden()

        for key, value in body.items():
            if key in NON_QUOTA_KEYS or (not value and value != 0):
                continue
            # validate whether already used and reserved exceeds the new
            # quota, this check will be ignored if admin want to force
            # update
            try:
                value = int(value)
            except (ValueError, TypeError):
                msg = _("Quota '%(value)s' for %(key)s should be "
                        "integer.") % {'value': value, 'key': key}
                LOG.warning(msg)
                raise webob.exc.HTTPBadRequest(explanation=msg)

            if force_update is False and value >= 0:
                quota_value = quotas.get(key)
                if quota_value and quota_value['limit'] >= 0:
                    quota_used = (quota_value['in_use'] +
                                  quota_value['reserved'])
                    LOG.debug("Quota %(key)s used: %(quota_used)s, "
                              "value: %(value)s.",
                              {'key': key, 'quota_used': quota_used,
                               'value': value})
                    if quota_used > value:
                        msg = (_("Quota value %(value)s for %(key)s is "
                                 "smaller than already used and reserved "
                                 "%(quota_used)s.") %
                               {'value': value, 'key': key,
                                'quota_used': quota_used})
                        raise webob.exc.HTTPBadRequest(explanation=msg)

            minimum = settable_quotas[key]['minimum']
            maximum = settable_quotas[key]['maximum']
            self._validate_quota_limit(value, minimum, maximum, force_update)
            try:
                db.quota_create(
                    context, project_id, key, value,
                    user_id=user_id, share_type_id=share_type_id)
            except exception.QuotaExists:
                db.quota_update(
                    context, project_id, key, value,
                    user_id=user_id, share_type_id=share_type_id)
            except exception.AdminRequired:
                raise webob.exc.HTTPForbidden()
        return self._view_builder.detail_list(
            req,
            self._get_quotas(
                context, id, user_id=user_id, share_type_id=share_type_id),
            share_type=share_type_id,
        )

    @wsgi.Controller.authorize("delete")
    def _delete(self, req, id):
        context = req.environ['manila.context']
        params = parse.parse_qs(req.environ.get('QUERY_STRING', ''))
        user_id = params.get('user_id', [None])[0]
        share_type = params.get('share_type', [None])[0]
        self._validate_user_id_and_share_type_args(user_id, share_type)
        try:
            db.authorize_project_context(context, id)
            if user_id:
                QUOTAS.destroy_all_by_project_and_user(context, id, user_id)
            elif share_type:
                share_type_id = self._get_share_type_id(context, share_type)
                QUOTAS.destroy_all_by_project_and_share_type(
                    context, id, share_type_id)
            else:
                QUOTAS.destroy_all_by_project(context, id)
            return webob.Response(status_int=http_client.ACCEPTED)
        except exception.NotAuthorized:
            raise webob.exc.HTTPForbidden()


class QuotaSetsControllerLegacy(QuotaSetsMixin, wsgi.Controller):
    """Deprecated Quota Sets API controller.

    Used by legacy API v1 and v2 microversions from 2.0 to 2.6.
    Registered under deprecated API URL 'os-quota-sets'.
    """

    @wsgi.Controller.api_version('1.0', '2.6')
    def show(self, req, id):
        self._ensure_share_type_arg_is_absent(req)
        return self._show(req, id)

    @wsgi.Controller.api_version('1.0', '2.6')
    def defaults(self, req, id):
        return self._defaults(req, id)

    @wsgi.Controller.api_version('1.0', '2.6')
    def update(self, req, id, body):
        self._ensure_share_type_arg_is_absent(req)
        self._ensure_specific_microversion_args_are_absent(
            body, ['share_groups', 'share_group_snapshots'], "2.40")
        self._ensure_specific_microversion_args_are_absent(
            body, ['share_replicas', 'replica_gigabytes'], "2.53")
        return self._update(req, id, body)

    @wsgi.Controller.api_version('1.0', '2.6')
    def delete(self, req, id):
        self._ensure_share_type_arg_is_absent(req)
        return self._delete(req, id)


class QuotaSetsController(QuotaSetsMixin, wsgi.Controller):
    """Quota Sets API controller.

    Used only by API v2 starting from microversion 2.7.
    Registered under API URL 'quota-sets'.
    """

    @wsgi.Controller.api_version('2.7')
    def show(self, req, id):
        if req.api_version_request < api_version.APIVersionRequest("2.39"):
            self._ensure_share_type_arg_is_absent(req)
        return self._show(req, id)

    @wsgi.Controller.api_version('2.25')
    def detail(self, req, id):
        if req.api_version_request < api_version.APIVersionRequest("2.39"):
            self._ensure_share_type_arg_is_absent(req)
        return self._show(req, id, True)

    @wsgi.Controller.api_version('2.7')
    def defaults(self, req, id):
        return self._defaults(req, id)

    @wsgi.Controller.api_version('2.7')
    def update(self, req, id, body):
        if req.api_version_request < api_version.APIVersionRequest("2.39"):
            self._ensure_share_type_arg_is_absent(req)
        elif req.api_version_request < api_version.APIVersionRequest("2.40"):
            self._ensure_specific_microversion_args_are_absent(
                body, ['share_groups', 'share_group_snapshots'], "2.40")
        elif req.api_version_request < api_version.APIVersionRequest("2.53"):
            self._ensure_specific_microversion_args_are_absent(
                body, ['share_replicas', 'replica_gigabytes'], "2.53")
        elif req.api_version_request < api_version.APIVersionRequest("2.62"):
            self._ensure_specific_microversion_args_are_absent(
                body, ['per_share_gigabytes'], "2.62")
        return self._update(req, id, body)

    @wsgi.Controller.api_version('2.7')
    def delete(self, req, id):
        if req.api_version_request < api_version.APIVersionRequest("2.39"):
            self._ensure_share_type_arg_is_absent(req)
        return self._delete(req, id)


def create_resource_legacy():
    return wsgi.Resource(QuotaSetsControllerLegacy())


def create_resource():
    return wsgi.Resource(QuotaSetsController())



# Copyright 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinderclient import exceptions as cinder_exception
import ddt

from manila import context
from manila import exception
from manila import test
from manila.tests import utils as test_utils
from manila.volume import cinder


class FakeCinderClient(object):
    class Volumes(object):
        def get(self, volume_id):
            return {'id': volume_id}

        def list(self, detailed, search_opts={}):
            return [{'id': 'id1'}, {'id': 'id2'}]

        def create(self, *args, **kwargs):
            return {'id': 'created_id'}

        def __getattr__(self, item):
            return None

    def __init__(self):
        self.volumes = self.Volumes()
        self.volume_snapshots = self.volumes


def get_fake_auth_obj():
    return type('FakeAuthObj', (object, ), {'get_client': mock.Mock()})


class CinderclientTestCase(test.TestCase):

    @mock.patch('manila.volume.cinder.AUTH_OBJ', None)
    def test_no_auth_obj(self):
        mock_client_loader = self.mock_object(
            cinder.client_auth, 'AuthClientLoader')
        fake_context = 'fake_context'
        data = {
            'cinder': {
                'http_retries': 3,
                'endpoint_type': 'foo_endpoint_type',
                'region_name': 'foo_region_name',
            }
        }

        with test_utils.create_temp_config_with_opts(data):
            cinder.cinderclient(fake_context)

        mock_client_loader.assert_called_once_with(
            client_class=cinder.cinder_client.Client,
            cfg_group=cinder.CINDER_GROUP
        )
        mock_client_loader.return_value.get_client.assert_called_once_with(
            fake_context,
            retries=data['cinder']['http_retries'],
            endpoint_type=data['cinder']['endpoint_type'],
            region_name=data['cinder']['region_name'],
        )

    @mock.patch('manila.volume.cinder.AUTH_OBJ', get_fake_auth_obj())
    def test_with_auth_obj(self):
        fake_context = 'fake_context'
        data = {
            'cinder': {
                'http_retries': 3,
                'endpoint_type': 'foo_endpoint_type',
                'region_name': 'foo_region_name',
            }
        }

        with test_utils.create_temp_config_with_opts(data):
            cinder.cinderclient(fake_context)

        cinder.AUTH_OBJ.get_client.assert_called_once_with(
            fake_context,
            retries=data['cinder']['http_retries'],
            endpoint_type=data['cinder']['endpoint_type'],
            region_name=data['cinder']['region_name'],
        )


@ddt.ddt
class CinderApiTestCase(test.TestCase):
    def setUp(self):
        super(CinderApiTestCase, self).setUp()

        self.api = cinder.API()
        self.cinderclient = FakeCinderClient()
        self.ctx = context.get_admin_context()
        self.mock_object(cinder, 'cinderclient',
                         mock.Mock(return_value=self.cinderclient))
        self.mock_object(cinder, '_untranslate_volume_summary_view',
                         lambda ctx, vol: vol)
        self.mock_object(cinder, '_untranslate_snapshot_summary_view',
                         lambda ctx, snap: snap)

    def test_get(self):
        volume_id = 'volume_id1'
        result = self.api.get(self.ctx, volume_id)
        self.assertEqual(volume_id, result['id'])

    @ddt.data(
        {'cinder_e': cinder_exception.NotFound(404),
         'manila_e': exception.VolumeNotFound},
        {'cinder_e': cinder_exception.BadRequest(400),
         'manila_e': exception.InvalidInput},
    )
    @ddt.unpack
    def test_get_failed(self, cinder_e, manila_e):
        cinder.cinderclient.side_effect = cinder_e
        volume_id = 'volume_id'
        self.assertRaises(manila_e, self.api.get, self.ctx, volume_id)

    def test_create(self):
        result = self.api.create(self.ctx, 1, '', '')
        self.assertEqual('created_id', result['id'])

    def test_create_failed(self):
        cinder.cinderclient.side_effect = cinder_exception.BadRequest(400)
        self.assertRaises(exception.InvalidInput,
                          self.api.create, self.ctx, 1, '', '')

    def test_create_not_found_error(self):
        cinder.cinderclient.side_effect = cinder_exception.NotFound(404)
        self.assertRaises(exception.NotFound,
                          self.api.create, self.ctx, 1, '', '')

    def test_create_failed_exception(self):
        cinder.cinderclient.side_effect = Exception("error msg")
        self.assertRaises(exception.ManilaException,
                          self.api.create, self.ctx, 1, '', '')

    def test_get_all(self):
        cinder._untranslate_volume_summary_view.return_value = ['id1', 'id2']
        self.assertEqual([{'id': 'id1'}, {'id': 'id2'}],
                         self.api.get_all(self.ctx))

    def test_check_attach_volume_status_error(self):
        volume = {'status': 'error'}
        self.assertRaises(exception.InvalidVolume,
                          self.api.check_attach, self.ctx, volume)

    def test_check_attach_volume_already_attached(self):
        volume = {'status': 'available'}
        volume['attach_status'] = "attached"
        self.assertRaises(exception.InvalidVolume,
                          self.api.check_attach, self.ctx, volume)

    def test_check_attach_availability_zone_differs(self):
        volume = {'status': 'available'}
        volume['attach_status'] = "detached"
        instance = {'availability_zone': 'zone1'}
        volume['availability_zone'] = 'zone2'
        cinder.CONF.set_override('cross_az_attach', False, 'cinder')
        self.assertRaises(exception.InvalidVolume,
                          self.api.check_attach, self.ctx, volume, instance)
        volume['availability_zone'] = 'zone1'
        self.assertIsNone(self.api.check_attach(self.ctx, volume, instance))
        cinder.CONF.reset()

    def test_check_attach(self):
        volume = {'status': 'available'}
        volume['attach_status'] = "detached"
        volume['availability_zone'] = 'zone1'
        instance = {'availability_zone': 'zone1'}
        cinder.CONF.set_override('cross_az_attach', False, 'cinder')
        self.assertIsNone(self.api.check_attach(self.ctx, volume, instance))
        cinder.CONF.reset()

    def test_check_detach(self):
        volume = {'status': 'available'}
        self.assertRaises(exception.InvalidVolume,
                          self.api.check_detach, self.ctx, volume)
        volume['status'] = 'non-available'
        self.assertIsNone(self.api.check_detach(self.ctx, volume))

    def test_update(self):
        fake_volume = {'fake': 'fake'}
        self.mock_object(self.cinderclient.volumes, 'get',
                         mock.Mock(return_value=fake_volume))
        self.mock_object(self.cinderclient.volumes, 'update')
        fake_volume_id = 'fake_volume'
        fake_data = {'test': 'test'}

        self.api.update(self.ctx, fake_volume_id, fake_data)

        self.cinderclient.volumes.get.assert_called_once_with(fake_volume_id)
        self.cinderclient.volumes.update.assert_called_once_with(fake_volume,
                                                                 **fake_data)

    def test_reserve_volume(self):
        self.mock_object(self.cinderclient.volumes, 'reserve')
        self.api.reserve_volume(self.ctx, 'id1')
        self.cinderclient.volumes.reserve.assert_called_once_with('id1')

    def test_unreserve_volume(self):
        self.mock_object(self.cinderclient.volumes, 'unreserve')
        self.api.unreserve_volume(self.ctx, 'id1')
        self.cinderclient.volumes.unreserve.assert_called_once_with('id1')

    def test_begin_detaching(self):
        self.mock_object(self.cinderclient.volumes, 'begin_detaching')
        self.api.begin_detaching(self.ctx, 'id1')
        self.cinderclient.volumes.begin_detaching.assert_called_once_with(
            'id1')

    def test_roll_detaching(self):
        self.mock_object(self.cinderclient.volumes, 'roll_detaching')
        self.api.roll_detaching(self.ctx, 'id1')
        self.cinderclient.volumes.roll_detaching.assert_called_once_with('id1')

    def test_attach(self):
        self.mock_object(self.cinderclient.volumes, 'attach')
        self.api.attach(self.ctx, 'id1', 'uuid', 'point')
        self.cinderclient.volumes.attach.assert_called_once_with('id1',
                                                                 'uuid',
                                                                 'point')

    def test_detach(self):
        self.mock_object(self.cinderclient.volumes, 'detach')
        self.api.detach(self.ctx, 'id1')
        self.cinderclient.volumes.detach.assert_called_once_with('id1')

    def test_initialize_connection(self):
        self.mock_object(self.cinderclient.volumes, 'initialize_connection')
        self.api.initialize_connection(self.ctx, 'id1', 'connector')
        (self.cinderclient.volumes.initialize_connection.
         assert_called_once_with('id1', 'connector'))

    def test_terminate_connection(self):
        self.mock_object(self.cinderclient.volumes, 'terminate_connection')
        self.api.terminate_connection(self.ctx, 'id1', 'connector')
        (self.cinderclient.volumes.terminate_connection.
         assert_called_once_with('id1', 'connector'))

    def test_delete(self):
        self.mock_object(self.cinderclient.volumes, 'delete')
        self.api.delete(self.ctx, 'id1')
        self.cinderclient.volumes.delete.assert_called_once_with('id1')

    def test_get_snapshot(self):
        snapshot_id = 'snapshot_id1'
        result = self.api.get_snapshot(self.ctx, snapshot_id)
        self.assertEqual(snapshot_id, result['id'])

    def test_get_snapshot_failed(self):
        cinder.cinderclient.side_effect = cinder_exception.NotFound(404)
        snapshot_id = 'snapshot_id'
        self.assertRaises(exception.VolumeSnapshotNotFound,
                          self.api.get_snapshot, self.ctx, snapshot_id)

    def test_get_all_snapshots(self):
        cinder._untranslate_snapshot_summary_view.return_value = ['id1', 'id2']
        self.assertEqual([{'id': 'id1'}, {'id': 'id2'}],
                         self.api.get_all_snapshots(self.ctx))

    def test_create_snapshot(self):
        result = self.api.create_snapshot(self.ctx, {'id': 'id1'}, '', '')
        self.assertEqual('created_id', result['id'])

    def test_create_force(self):
        result = self.api.create_snapshot_force(self.ctx,
                                                {'id': 'id1'}, '', '')
        self.assertEqual('created_id', result['id'])

    def test_delete_snapshot(self):
        self.mock_object(self.cinderclient.volume_snapshots, 'delete')
        self.api.delete_snapshot(self.ctx, 'id1')
        self.cinderclient.volume_snapshots.delete.assert_called_once_with(
            'id1')



# Copyright (c) 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from unittest import mock

import ddt
from oslo_config import cfg

from manila import exception
from manila.share.drivers.ganesha import utils as ganesha_utils
from manila.share.drivers.zfsonlinux import utils as zfs_utils
from manila import test

CONF = cfg.CONF


def get_fake_configuration(*args, **kwargs):
    fake_config_options = {
        "zfs_use_ssh": kwargs.get("zfs_use_ssh", False),
        "zfs_share_export_ip": kwargs.get(
            "zfs_share_export_ip", "240.241.242.243"),
        "zfs_service_ip": kwargs.get("zfs_service_ip", "240.241.242.244"),
        "ssh_conn_timeout": kwargs.get("ssh_conn_timeout", 123),
        "zfs_ssh_username": kwargs.get(
            "zfs_ssh_username", 'fake_username'),
        "zfs_ssh_user_password": kwargs.get(
            "zfs_ssh_user_password", 'fake_pass'),
        "zfs_ssh_private_key_path": kwargs.get(
            "zfs_ssh_private_key_path", '/fake/path'),
        "append_config_values": mock.Mock(),
    }
    return type("FakeConfig", (object, ), fake_config_options)


class FakeShareDriver(zfs_utils.ExecuteMixin):
    def __init__(self, *args, **kwargs):
        self.configuration = get_fake_configuration(*args, **kwargs)
        self.init_execute_mixin(*args, **kwargs)


@ddt.ddt
class ExecuteMixinTestCase(test.TestCase):

    def setUp(self):
        super(ExecuteMixinTestCase, self).setUp()
        self.ssh_executor = self.mock_object(ganesha_utils, 'SSHExecutor')
        self.driver = FakeShareDriver()

    def test_init(self):
        self.assertIsNone(self.driver.ssh_executor)
        self.assertEqual(0, self.ssh_executor.call_count)

    def test_init_ssh(self):
        driver = FakeShareDriver(zfs_use_ssh=True)

        self.assertIsNotNone(driver.ssh_executor)
        self.ssh_executor.assert_called_once_with(
            ip=driver.configuration.zfs_service_ip,
            port=22,
            conn_timeout=driver.configuration.ssh_conn_timeout,
            login=driver.configuration.zfs_ssh_username,
            password=driver.configuration.zfs_ssh_user_password,
            privatekey=driver.configuration.zfs_ssh_private_key_path,
            max_size=10,
        )

    def test_execute_with_provided_executor(self):
        self.mock_object(self.driver, '_execute')
        fake_executor = mock.Mock()

        self.driver.execute('fake', '--foo', '--bar', executor=fake_executor)

        self.assertFalse(self.driver._execute.called)
        self.assertFalse(self.ssh_executor.called)
        fake_executor.assert_called_once_with('fake', '--foo', '--bar')

    def test_local_shell_execute(self):
        self.mock_object(self.driver, '_execute')

        self.driver.execute('fake', '--foo', '--bar')

        self.assertEqual(0, self.ssh_executor.call_count)
        self.driver._execute.assert_called_once_with(
            'fake', '--foo', '--bar')

    def test_local_shell_execute_with_sudo(self):
        self.mock_object(self.driver, '_execute')

        self.driver.execute('sudo', 'fake', '--foo', '--bar')

        self.assertEqual(0, self.ssh_executor.call_count)
        self.driver._execute.assert_called_once_with(
            'fake', '--foo', '--bar', run_as_root=True)

    def test_ssh_execute(self):
        driver = FakeShareDriver(zfs_use_ssh=True)

        self.mock_object(driver, '_execute')

        driver.execute('fake', '--foo', '--bar')

        self.assertEqual(0, driver._execute.call_count)
        self.ssh_executor.return_value.assert_called_once_with(
            'fake', '--foo', '--bar')

    def test_ssh_execute_with_sudo(self):
        driver = FakeShareDriver(zfs_use_ssh=True)

        self.mock_object(driver, '_execute')

        driver.execute('sudo', 'fake', '--foo', '--bar')

        self.assertEqual(0, driver._execute.call_count)
        self.ssh_executor.return_value.assert_called_once_with(
            'fake', '--foo', '--bar', run_as_root=True)

    def test_execute_with_retry(self):
        self.mock_object(time, 'sleep')
        self.mock_object(self.driver, 'execute', mock.Mock(
            side_effect=[exception.ProcessExecutionError('FAKE'), None]))
        self.driver.execute_with_retry('foo', 'bar')

        self.assertEqual(2, self.driver.execute.call_count)
        self.driver.execute.assert_has_calls(
            [mock.call('foo', 'bar'), mock.call('foo', 'bar')])

    def test_execute_with_retry_exceeded(self):
        self.mock_object(time, 'sleep')
        self.mock_object(self.driver, 'execute', mock.Mock(
            side_effect=exception.ProcessExecutionError('FAKE')))

        self.assertRaises(
            exception.ProcessExecutionError,
            self.driver.execute_with_retry,
            'foo', 'bar',
        )

        self.assertEqual(36, self.driver.execute.call_count)

    @ddt.data(True, False)
    def test__get_option(self, pool_level):
        out = """NAME   PROPERTY              VALUE                  SOURCE\n
foo_resource_name  bar_option_name           some_value              local"""
        self.mock_object(
            self.driver, '_execute', mock.Mock(return_value=(out, '')))
        res_name = 'foo_resource_name'
        opt_name = 'bar_option_name'

        result = self.driver._get_option(
            res_name, opt_name, pool_level=pool_level)

        self.assertEqual('some_value', result)
        self.driver._execute.assert_called_once_with(
            'zpool' if pool_level else 'zfs', 'get', opt_name, res_name,
            run_as_root=True)

    def test_parse_zfs_answer(self):
        not_parsed_str = ''
        not_parsed_str = """NAME   PROPERTY       VALUE              SOURCE\n
foo_res  opt_1           bar              local
foo_res  opt_2           foo              default
foo_res  opt_3           some_value       local"""
        expected = [
            {'NAME': 'foo_res', 'PROPERTY': 'opt_1', 'VALUE': 'bar',
             'SOURCE': 'local'},
            {'NAME': 'foo_res', 'PROPERTY': 'opt_2', 'VALUE': 'foo',
             'SOURCE': 'default'},
            {'NAME': 'foo_res', 'PROPERTY': 'opt_3', 'VALUE': 'some_value',
             'SOURCE': 'local'},
        ]

        result = self.driver.parse_zfs_answer(not_parsed_str)

        self.assertEqual(expected, result)

    def test_parse_zfs_answer_empty(self):
        result = self.driver.parse_zfs_answer('')

        self.assertEqual([], result)

    def test_get_zpool_option(self):
        self.mock_object(self.driver, '_get_option')
        zpool_name = 'foo_resource_name'
        opt_name = 'bar_option_name'

        result = self.driver.get_zpool_option(zpool_name, opt_name)

        self.assertEqual(self.driver._get_option.return_value, result)
        self.driver._get_option.assert_called_once_with(
            zpool_name, opt_name, True)

    def test_get_zfs_option(self):
        self.mock_object(self.driver, '_get_option')
        dataset_name = 'foo_resource_name'
        opt_name = 'bar_option_name'

        result = self.driver.get_zfs_option(dataset_name, opt_name)

        self.assertEqual(self.driver._get_option.return_value, result)
        self.driver._get_option.assert_called_once_with(
            dataset_name, opt_name, False)

    def test_zfs(self):
        self.mock_object(self.driver, 'execute')
        self.mock_object(self.driver, 'execute_with_retry')

        self.driver.zfs('foo', 'bar')

        self.assertEqual(0, self.driver.execute_with_retry.call_count)
        self.driver.execute.assert_called_once_with(
            'sudo', 'zfs', 'foo', 'bar')

    def test_zfs_with_retry(self):
        self.mock_object(self.driver, 'execute')
        self.mock_object(self.driver, 'execute_with_retry')

        self.driver.zfs_with_retry('foo', 'bar')

        self.assertEqual(0, self.driver.execute.call_count)
        self.driver.execute_with_retry.assert_called_once_with(
            'sudo', 'zfs', 'foo', 'bar')


@ddt.ddt
class NFSviaZFSHelperTestCase(test.TestCase):

    def setUp(self):
        super(NFSviaZFSHelperTestCase, self).setUp()
        configuration = get_fake_configuration()
        self.out = "fake_out"
        self.mock_object(
            zfs_utils.utils, "execute", mock.Mock(return_value=(self.out, "")))
        self.helper = zfs_utils.NFSviaZFSHelper(configuration)

    def test_init(self):
        zfs_utils.utils.execute.assert_has_calls([
            mock.call("which", "exportfs"),
            mock.call("exportfs", run_as_root=True),
        ])

    def test_verify_setup_exportfs_not_installed(self):
        zfs_utils.utils.execute.reset_mock()
        zfs_utils.utils.execute.side_effect = [('', '')]

        self.assertRaises(
            exception.ZFSonLinuxException, self.helper.verify_setup)

        zfs_utils.utils.execute.assert_called_once_with("which", "exportfs")

    def test_verify_setup_error_calling_exportfs(self):
        zfs_utils.utils.execute.reset_mock()
        zfs_utils.utils.execute.side_effect = [
            ('fake_out', ''), exception.ProcessExecutionError('Fake')]

        self.assertRaises(
            exception.ProcessExecutionError, self.helper.verify_setup)

        zfs_utils.utils.execute.assert_has_calls([
            mock.call("which", "exportfs"),
            mock.call("exportfs", run_as_root=True),
        ])

    def test_is_kernel_version_true(self):
        delattr(self.helper, '_is_kernel_version')
        zfs_utils.utils.execute.reset_mock()

        self.assertTrue(self.helper.is_kernel_version)

        zfs_utils.utils.execute.assert_has_calls([
            mock.call("modinfo", "zfs"),
        ])

    def test_is_kernel_version_false(self):
        delattr(self.helper, '_is_kernel_version')
        zfs_utils.utils.execute.reset_mock()
        zfs_utils.utils.execute.side_effect = (
            exception.ProcessExecutionError('Fake'))

        self.assertFalse(self.helper.is_kernel_version)

        zfs_utils.utils.execute.assert_has_calls([
            mock.call("modinfo", "zfs"),
        ])

    def test_is_kernel_version_second_call(self):
        delattr(self.helper, '_is_kernel_version')
        zfs_utils.utils.execute.reset_mock()

        self.assertTrue(self.helper.is_kernel_version)
        self.assertTrue(self.helper.is_kernel_version)

        zfs_utils.utils.execute.assert_has_calls([
            mock.call("modinfo", "zfs"),
        ])

    def test_create_exports(self):
        self.mock_object(self.helper, 'get_exports')

        result = self.helper.create_exports('foo')

        self.assertEqual(
            self.helper.get_exports.return_value, result)

    def test_get_exports(self):
        self.mock_object(
            self.helper, 'get_zfs_option', mock.Mock(return_value='fake_mp'))
        expected = [
            {
                "path": "%s:fake_mp" % ip,
                "metadata": {},
                "is_admin_only": is_admin_only,
            } for ip, is_admin_only in (
                (self.helper.configuration.zfs_share_export_ip, False),
                (self.helper.configuration.zfs_service_ip, True))
        ]

        result = self.helper.get_exports('foo')

        self.assertEqual(expected, result)
        self.helper.get_zfs_option.assert_called_once_with(
            'foo', 'mountpoint', executor=None)

    def test_remove_exports(self):
        zfs_utils.utils.execute.reset_mock()
        self.mock_object(
            self.helper, 'get_zfs_option', mock.Mock(return_value='bar'))

        self.helper.remove_exports('foo')

        self.helper.get_zfs_option.assert_called_once_with(
            'foo', 'sharenfs', executor=None)
        zfs_utils.utils.execute.assert_called_once_with(
            'zfs', 'set', 'sharenfs=off', 'foo', run_as_root=True)

    def test_remove_exports_that_absent(self):
        zfs_utils.utils.execute.reset_mock()
        self.mock_object(
            self.helper, 'get_zfs_option', mock.Mock(return_value='off'))

        self.helper.remove_exports('foo')

        self.helper.get_zfs_option.assert_called_once_with(
            'foo', 'sharenfs', executor=None)
        self.assertEqual(0, zfs_utils.utils.execute.call_count)

    @ddt.data(
        (('fake_modinfo_result', ''),
         ('sharenfs=rw=1.1.1.1:3.3.3.0/255.255.255.0,no_root_squash,'
          'ro=2.2.2.2,no_root_squash'), False),
        (('fake_modinfo_result', ''),
         ('sharenfs=ro=1.1.1.1:2.2.2.2:3.3.3.0/255.255.255.0,no_root_squash'),
         True),
        (exception.ProcessExecutionError('Fake'),
         ('sharenfs=1.1.1.1:rw,no_root_squash 3.3.3.0/255.255.255.0:rw,'
          'no_root_squash 2.2.2.2:ro,no_root_squash'), False),
        (exception.ProcessExecutionError('Fake'),
         ('sharenfs=1.1.1.1:ro,no_root_squash 2.2.2.2:ro,'
          'no_root_squash 3.3.3.0/255.255.255.0:ro,no_root_squash'), True),
    )
    @ddt.unpack
    def test_update_access_rw_and_ro(self, modinfo_response, access_str,
                                     make_all_ro):
        delattr(self.helper, '_is_kernel_version')
        zfs_utils.utils.execute.reset_mock()
        dataset_name = 'zpoolz/foo_dataset_name/fake'
        zfs_utils.utils.execute.side_effect = [
            modinfo_response,
            ("""NAME            USED  AVAIL  REFER  MOUNTPOINT\n
%(dn)s                2.58M  14.8G  27.5K  /%(dn)s\n
%(dn)s_some_other     3.58M  15.8G  28.5K  /%(dn)s\n
             """ % {'dn': dataset_name}, ''),
            ('fake_set_opt_result', ''),
            ("""NAME                     PROPERTY    VALUE            SOURCE\n
%s          mountpoint  /%s  default\n
             """ % (dataset_name, dataset_name), ''),
            ('fake_1_result', ''),
            ('fake_2_result', ''),
            ('fake_3_result', ''),
            ('fake_4_result', ''),
            ('fake_5_result', ''),
        ]
        access_rules = [
            {'access_type': 'ip', 'access_level': 'rw',
             'access_to': '1.1.1.1'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '2.2.2.2'},
            {'access_type': 'ip', 'access_level': 'rw',
             'access_to': '3.3.3.0/24'},
        ]
        delete_rules = [
            {'access_type': 'ip', 'access_level': 'rw',
             'access_to': '4.4.4.4'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '5.5.5.5/32'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '5.5.5.6/16'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '5.5.5.7/0'},
            {'access_type': 'user', 'access_level': 'rw',
             'access_to': '6.6.6.6'},
            {'access_type': 'user', 'access_level': 'ro',
             'access_to': '7.7.7.7'},
        ]

        self.helper.update_access(
            dataset_name, access_rules, [], delete_rules,
            make_all_ro=make_all_ro)

        zfs_utils.utils.execute.assert_has_calls([
            mock.call('modinfo', 'zfs'),
            mock.call('zfs', 'list', '-r', 'zpoolz', run_as_root=True),
            mock.call(
                'zfs', 'set',
                access_str,
                dataset_name, run_as_root=True),
            mock.call(
                'zfs', 'get', 'mountpoint', dataset_name, run_as_root=True),
            mock.call(
                'exportfs', '-u', '4.4.4.4:/%s' % dataset_name,
                run_as_root=True),
            mock.call(
                'exportfs', '-u', '5.5.5.5:/%s' % dataset_name,
                run_as_root=True),
            mock.call(
                'exportfs', '-u', '5.5.5.6/255.255.0.0:/%s' % dataset_name,
                run_as_root=True),
            mock.call(
                'exportfs', '-u', '5.5.5.7/0.0.0.0:/%s' % dataset_name,
                run_as_root=True),
        ])

    def test_update_access_dataset_not_found(self):
        self.mock_object(zfs_utils.LOG, 'warning')
        zfs_utils.utils.execute.reset_mock()
        dataset_name = 'zpoolz/foo_dataset_name/fake'
        zfs_utils.utils.execute.side_effect = [
            ('fake_modinfo_result', ''),
            ('fake_dataset_not_found_result', ''),
            ('fake_set_opt_result', ''),
        ]
        access_rules = [
            {'access_type': 'ip', 'access_level': 'rw',
             'access_to': '1.1.1.1'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '1.1.1.2'},
        ]

        self.helper.update_access(dataset_name, access_rules, [], [])

        zfs_utils.utils.execute.assert_has_calls([
            mock.call('zfs', 'list', '-r', 'zpoolz', run_as_root=True),
        ])
        zfs_utils.LOG.warning.assert_called_once_with(
            mock.ANY, {'name': dataset_name})

    @ddt.data(exception.ProcessExecutionError('Fake'), ('Ok', ''))
    def test_update_access_no_rules(self, first_execute_result):
        zfs_utils.utils.execute.reset_mock()
        dataset_name = 'zpoolz/foo_dataset_name/fake'
        zfs_utils.utils.execute.side_effect = [
            ("""NAME            USED  AVAIL  REFER  MOUNTPOINT\n
%s          2.58M  14.8G  27.5K  /%s\n
             """ % (dataset_name, dataset_name), ''),
            ('fake_set_opt_result', ''),
        ]

        self.helper.update_access(dataset_name, [], [], [])

        zfs_utils.utils.execute.assert_has_calls([
            mock.call('zfs', 'list', '-r', 'zpoolz', run_as_root=True),
            mock.call('zfs', 'set', 'sharenfs=off', dataset_name,
                      run_as_root=True),
        ])

    @ddt.data('user', 'cert', 'cephx', '', 'fake', 'i', 'p')
    def test_update_access_not_ip_access_type(self, access_type):
        zfs_utils.utils.execute.reset_mock()
        dataset_name = 'zpoolz/foo_dataset_name/fake'
        access_rules = [
            {'access_type': access_type, 'access_level': 'rw',
             'access_to': '1.1.1.1'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '1.1.1.2'},
        ]

        self.assertRaises(
            exception.InvalidShareAccess,
            self.helper.update_access,
            dataset_name, access_rules, access_rules, [],
        )

        self.assertEqual(0, zfs_utils.utils.execute.call_count)

    @ddt.data('', 'r', 'o', 'w', 'fake', 'su')
    def test_update_access_neither_rw_nor_ro_access_level(self, access_level):
        zfs_utils.utils.execute.reset_mock()
        dataset_name = 'zpoolz/foo_dataset_name/fake'
        access_rules = [
            {'access_type': 'ip', 'access_level': access_level,
             'access_to': '1.1.1.1'},
            {'access_type': 'ip', 'access_level': 'ro',
             'access_to': '1.1.1.2'},
        ]

        self.assertRaises(
            exception.InvalidShareAccess,
            self.helper.update_access,
            dataset_name, access_rules, access_rules, [],
        )

        self.assertEqual(0, zfs_utils.utils.execute.call_count)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Test of Share Type methods for Manila."""
import copy
import datetime
from unittest import mock

import ddt

from manila.common import constants
from manila import context
from manila import db
from manila import exception
from manila.share_group import share_group_types
from manila import test


def create_share_group_type_dict(group_specs=None):
    return {
        'fake_type': {
            'name': 'fake1',
            'group_specs': group_specs
        }
    }


@ddt.ddt
class ShareGroupTypesTestCase(test.TestCase):

    fake_type = {
        'test': {
            'created_at': datetime.datetime(2015, 1, 22, 11, 43, 24),
            'deleted': '0',
            'deleted_at': None,
            'group_specs': {},
            'id': u'fooid-1',
            'name': u'test',
            'updated_at': None
        }
    }
    fake_group_specs = {u'gold': u'True'}
    fake_share_group_type_id = u'fooid-2'
    fake_type_w_extra = {
        'test_with_extra': {
            'created_at': datetime.datetime(2015, 1, 22, 11, 45, 31),
            'deleted': '0',
            'deleted_at': None,
            'group_specs': fake_group_specs,
            'id': fake_share_group_type_id,
            'name': u'test_with_extra',
            'updated_at': None
        }
    }

    fake_type_w_valid_extra = {
        'test_with_extra': {
            'created_at': datetime.datetime(2015, 1, 22, 11, 45, 31),
            'deleted': '0',
            'deleted_at': None,
            'group_specs': {
                constants.ExtraSpecs.DRIVER_HANDLES_SHARE_SERVERS: 'true'
            },
            'id': u'fooid-2',
            'name': u'test_with_extra',
            'updated_at': None
        }
    }

    fake_types = fake_type.copy()
    fake_types.update(fake_type_w_extra)
    fake_types.update(fake_type_w_valid_extra)

    fake_share_group = {
        'id': u'fooid-1',
        'share_group_type_id': fake_share_group_type_id,
    }

    def setUp(self):
        super(ShareGroupTypesTestCase, self).setUp()
        self.context = context.get_admin_context()

    @ddt.data({}, fake_type, fake_type_w_extra, fake_types)
    def test_get_all_types(self, share_group_type):
        self.mock_object(
            db, 'share_group_type_get_all',
            mock.Mock(return_value=copy.deepcopy(share_group_type)))

        returned_type = share_group_types.get_all(self.context)

        self.assertEqual(sorted(share_group_type), sorted(returned_type))

    def test_get_all_types_search(self):
        share_group_type = self.fake_type_w_extra
        search_filter = {"group_specs": {"gold": "True"}, 'is_public': True}
        self.mock_object(
            db, 'share_group_type_get_all',
            mock.Mock(return_value=share_group_type))

        returned_type = share_group_types.get_all(
            self.context, search_opts=search_filter)

        db.share_group_type_get_all.assert_called_once_with(
            mock.ANY, 0, filters={'is_public': True})
        self.assertEqual(sorted(share_group_type), sorted(returned_type))
        search_filter = {"group_specs": {"gold": "False"}}
        returned_type = share_group_types.get_all(
            self.context, search_opts=search_filter)
        self.assertEqual({}, returned_type)

    def test_add_access(self):
        project_id = '456'
        share_group_type = share_group_types.create(self.context, 'type2', [])
        share_group_type_id = share_group_type.get('id')

        share_group_types.add_share_group_type_access(
            self.context, share_group_type_id, project_id)
        stype_access = db.share_group_type_access_get_all(
            self.context, share_group_type_id)

        self.assertIn(project_id, [a.project_id for a in stype_access])

    def test_add_access_invalid(self):
        self.assertRaises(
            exception.InvalidShareGroupType,
            share_group_types.add_share_group_type_access,
            'fake', None, 'fake')

    def test_remove_access(self):
        project_id = '456'
        share_group_type = share_group_types.create(
            self.context, 'type3', [], projects=['456'])
        share_group_type_id = share_group_type.get('id')

        share_group_types.remove_share_group_type_access(
            self.context, share_group_type_id, project_id)
        stype_access = db.share_group_type_access_get_all(
            self.context, share_group_type_id)

        self.assertNotIn(project_id, stype_access)

    def test_remove_access_invalid(self):
        self.assertRaises(
            exception.InvalidShareGroupType,
            share_group_types.remove_share_group_type_access,
            'fake', None, 'fake')



# Copyright (c) 2011-2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Pluggable Weighing support
"""

import abc

from manila.scheduler import base_handler


def normalize(weight_list, minval=None, maxval=None):
    """Normalize the values in a list between 0 and 1.0.

    The normalization is made regarding the lower and upper values present in
    weight_list. If the minval and/or maxval parameters are set, these values
    will be used instead of the minimum and maximum from the list.

    If all the values are equal, they are normalized to 0.
    """

    if not weight_list:
        return ()

    if maxval is None:
        maxval = max(weight_list)

    if minval is None:
        minval = min(weight_list)

    maxval = float(maxval)
    minval = float(minval)

    if minval == maxval:
        return [0] * len(weight_list)

    range_ = maxval - minval
    return ((i - minval) / range_ for i in weight_list)


class WeighedObject(object):
    """Object with weight information."""
    def __init__(self, obj, weight):
        self.obj = obj
        self.weight = weight

    def __repr__(self):
        return "<WeighedObject '%s': %s>" % (self.obj, self.weight)


class BaseWeigher(metaclass=abc.ABCMeta):
    """Base class for pluggable weighers.

    The attributes maxval and minval can be specified to set up the maximum
    and minimum values for the weighed objects. These values will then be
    taken into account in the normalization step, instead of taking the values
    from the calculated weighers.
    """

    minval = None
    maxval = None

    def weight_multiplier(self):
        """How weighted this weigher should be.

        Override this method in a subclass, so that the returned value is
        read from a configuration option to permit operators specify a
        multiplier for the weigher.
        """
        return 1.0

    @abc.abstractmethod
    def _weigh_object(self, obj, weight_properties):
        """Override in a subclass to specify a weight for a specific object."""

    def weigh_objects(self, weighed_obj_list, weight_properties):
        """Weigh multiple objects.

        Override in a subclass if you need access to all objects in order
        to calculate weighers. Do not modify the weight of an object here,
        just return a list of weighers.
        """
        # Calculate the weighers
        weights = []
        for obj in weighed_obj_list:
            weight = self._weigh_object(obj.obj, weight_properties)

            # Record the min and max values if they are None. If they anything
            # but none we assume that the weigher has set them
            if self.minval is None:
                self.minval = weight
            if self.maxval is None:
                self.maxval = weight

            if weight < self.minval:
                self.minval = weight
            elif weight > self.maxval:
                self.maxval = weight

            weights.append(weight)

        return weights


class BaseWeightHandler(base_handler.BaseHandler):
    object_class = WeighedObject

    def get_weighed_objects(self, weigher_classes, obj_list,
                            weighing_properties):
        """Return a sorted (descending), normalized list of WeighedObjects."""

        if not obj_list:
            return []

        weighed_objs = [self.object_class(obj, 0.0) for obj in obj_list]
        for weigher_cls in weigher_classes:
            weigher = weigher_cls()
            weights = weigher.weigh_objects(weighed_objs, weighing_properties)

            # Normalize the weighers
            weights = normalize(weights,
                                minval=weigher.minval,
                                maxval=weigher.maxval)

            for i, weight in enumerate(weights):
                obj = weighed_objs[i]
                obj.weight += weigher.weight_multiplier() * weight

        return sorted(weighed_objs, key=lambda x: x.weight, reverse=True)



# Copyright 2012 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from urllib import parse

import ddt
import webob

from manila.api.v1 import security_service
from manila.common import constants
from manila import db
from manila import exception
from manila import test
from manila.tests.api import fakes


@ddt.ddt
class ShareApiTest(test.TestCase):
    """Share Api Test."""
    def setUp(self):
        super(ShareApiTest, self).setUp()
        self.controller = security_service.SecurityServiceController()
        self.maxDiff = None
        self.ss_active_directory = {
            "created_at": "fake-time",
            "updated_at": "fake-time-2",
            "id": 1,
            "name": "fake-name",
            "description": "Fake Security Service Desc",
            "type": constants.SECURITY_SERVICES_ALLOWED_TYPES[0],
            "dns_ip": "1.1.1.1",
            "server": "fake-server",
            "domain": "fake-domain",
            "user": "fake-user",
            "password": "fake-password",
            "status": constants.STATUS_NEW,
            "project_id": "fake",
        }
        self.ss_ldap = {
            "created_at": "fake-time",
            "updated_at": "fake-time-2",
            "id": 2,
            "name": "ss-ldap",
            "description": "Fake Security Service Desc",
            "type": constants.SECURITY_SERVICES_ALLOWED_TYPES[1],
            "dns_ip": "2.2.2.2",
            "server": "test-server",
            "domain": "test-domain",
            "user": "test-user",
            "password": "test-password",
            "status": "active",
            "project_id": "fake",
        }
        self.valid_search_opts = {
            'user': 'fake-user',
            'server': 'fake-server',
            'dns_ip': '1.1.1.1',
            'domain': 'fake-domain',
            'type': constants.SECURITY_SERVICES_ALLOWED_TYPES[0],
        }
        self.check_policy_patcher = mock.patch(
            'manila.api.v1.security_service.policy.check_policy')
        self.check_policy_patcher.start()
        self.addCleanup(self._stop_started_patcher, self.check_policy_patcher)
        self.security_service_list_expected_resp = {
            'security_services': [{
                'id': self.ss_active_directory['id'],
                'name': self.ss_active_directory['name'],
                'type': self.ss_active_directory['type'],
                'status': self.ss_active_directory['status']
            }, ]
        }

        self.fake_share_network_list_with_share_servers = [{
            'id': 'fake_sn_id',
            'share_network_subnets': [{
                'id': 'fake_sns_id',
                'share_servers': [{'id': 'fake_ss_id'}]
            }]
        }]
        self.fake_share_network_list_without_share_servers = [{
            'id': 'fake_sn_id',
            'share_network_subnets': [{
                'id': 'fake_sns_id',
                'share_servers': []
            }]
        }]

    def _stop_started_patcher(self, patcher):
        if hasattr(patcher, 'is_local'):
            patcher.stop()

    def test_security_service_show(self):
        db.security_service_get = mock.Mock(
            return_value=self.ss_active_directory)
        req = fakes.HTTPRequest.blank('/security-services/1')
        res_dict = self.controller.show(req, '1')
        expected = self.ss_active_directory.copy()
        expected.update()
        self.assertEqual({'security_service': self.ss_active_directory},
                         res_dict)

    def test_security_service_show_not_found(self):
        db.security_service_get = mock.Mock(side_effect=exception.NotFound)
        req = fakes.HTTPRequest.blank('/shares/1')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.show,
                          req, '1')

    def test_security_service_create(self):
        sec_service = self.ss_active_directory.copy()
        create_stub = mock.Mock(
            return_value=sec_service)
        self.mock_object(db, 'security_service_create', create_stub)

        req = fakes.HTTPRequest.blank('/security-services')
        res_dict = self.controller.create(
            req, {"security_service": sec_service})
        expected = self.ss_active_directory.copy()
        self.assertEqual({'security_service': expected}, res_dict)

    def test_security_service_create_invalid_types(self):
        sec_service = self.ss_active_directory.copy()
        sec_service['type'] = 'invalid'
        req = fakes.HTTPRequest.blank('/security-services')
        self.assertRaises(exception.InvalidInput, self.controller.create, req,
                          {"security_service": sec_service})

    def test_create_security_service_no_body(self):
        body = {}
        req = fakes.HTTPRequest.blank('/security-services')
        self.assertRaises(webob.exc.HTTPUnprocessableEntity,
                          self.controller.create,
                          req,
                          body)

    def test_security_service_delete(self):
        db.security_service_delete = mock.Mock()
        db.security_service_get = mock.Mock()
        db.share_network_get_all_by_security_service = mock.Mock(
            return_value=[])
        req = fakes.HTTPRequest.blank('/security_services/1')
        resp = self.controller.delete(req, 1)
        db.security_service_delete.assert_called_once_with(
            req.environ['manila.context'], 1)
        self.assertEqual(202, resp.status_int)

    def test_security_service_delete_not_found(self):
        db.security_service_get = mock.Mock(side_effect=exception.NotFound)
        req = fakes.HTTPRequest.blank('/security_services/1')
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.delete,
                          req,
                          1)

    def test_security_service_delete_has_share_networks(self):
        db.security_service_get = mock.Mock()
        db.share_network_get_all_by_security_service = mock.Mock(
            return_value=[{'share_network': 'fake_share_network'}])
        req = fakes.HTTPRequest.blank('/security_services/1')
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.delete,
                          req, 1)

    def test_security_service_update_name(self):
        new = self.ss_active_directory.copy()
        updated = self.ss_active_directory.copy()
        updated['name'] = 'new'
        self.mock_object(security_service.policy, 'check_policy')
        db.security_service_get = mock.Mock(return_value=new)
        db.security_service_update = mock.Mock(return_value=updated)
        fake_sns = {'id': 'fake_sns_id', 'share_servers': ['fake_ss']}
        db.share_network_get_all_by_security_service = mock.Mock(
            return_value=[{
                'id': 'fake_id',
                'share_network_subnets': [fake_sns]
            }])
        body = {"security_service": {"name": "new"}}
        req = fakes.HTTPRequest.blank('/security_service/1')
        res_dict = self.controller.update(req, 1, body)['security_service']
        self.assertEqual(updated['name'], res_dict['name'])
        db.share_network_get_all_by_security_service.assert_called_once_with(
            req.environ['manila.context'], 1)
        self.assertEqual(2, security_service.policy.check_policy.call_count)
        security_service.policy.check_policy.assert_has_calls([
            mock.call(req.environ['manila.context'],
                      security_service.RESOURCE_NAME, 'update', new)
        ])

    def test_security_service_update_description(self):
        new = self.ss_active_directory.copy()
        updated = self.ss_active_directory.copy()
        updated['description'] = 'new'
        self.mock_object(security_service.policy, 'check_policy')
        db.security_service_get = mock.Mock(return_value=new)
        db.security_service_update = mock.Mock(return_value=updated)
        fake_sns = {'id': 'fake_sns_id', 'share_servers': ['fake_ss']}
        db.share_network_get_all_by_security_service = mock.Mock(
            return_value=[{
                'id': 'fake_id',
                'share_network_subnets': [fake_sns]
            }])
        body = {"security_service": {"description": "new"}}
        req = fakes.HTTPRequest.blank('/security_service/1')
        res_dict = self.controller.update(req, 1, body)['security_service']
        self.assertEqual(updated['description'], res_dict['description'])
        db.share_network_get_all_by_security_service.assert_called_once_with(
            req.environ['manila.context'], 1)
        self.assertEqual(2, security_service.policy.check_policy.call_count)
        security_service.policy.check_policy.assert_has_calls([
            mock.call(req.environ['manila.context'],
                      security_service.RESOURCE_NAME, 'update', new)
        ])

    @mock.patch.object(db, 'security_service_get', mock.Mock())
    @mock.patch.object(db, 'share_network_get_all_by_security_service',
                       mock.Mock())
    def test_security_service_update_invalid_keys_sh_server_exists(self):
        self.mock_object(security_service.policy, 'check_policy')
        fake_sns = {'id': 'fake_sns_id', 'share_servers': ['fake_ss']}
        db.share_network_get_all_by_security_service.return_value = [
            {'id': 'fake_id', 'share_network_subnets': [fake_sns]},
        ]
        db.security_service_get.return_value = self.ss_active_directory.copy()
        body = {'security_service': {'user_id': 'new_user'}}
        req = fakes.HTTPRequest.blank('/security_services/1')
        self.assertRaises(webob.exc.HTTPForbidden, self.controller.update,
                          req, 1, body)
        db.security_service_get.assert_called_once_with(
            req.environ['manila.context'], 1)
        db.share_network_get_all_by_security_service.assert_called_once_with(
            req.environ['manila.context'], 1)
        self.assertEqual(1, security_service.policy.check_policy.call_count)
        security_service.policy.check_policy.assert_has_calls([
            mock.call(req.environ['manila.context'],
                      security_service.RESOURCE_NAME, 'update',
                      db.security_service_get.return_value)
        ])

    @mock.patch.object(db, 'security_service_get', mock.Mock())
    @mock.patch.object(db, 'security_service_update', mock.Mock())
    @mock.patch.object(db, 'share_network_get_all_by_security_service',
                       mock.Mock())
    def test_security_service_update_valid_keys_sh_server_exists(self):
        self.mock_object(security_service.policy, 'check_policy')
        fake_sns = {'id': 'fake_sns_id', 'share_servers': ['fake_ss']}
        db.share_network_get_all_by_security_service.return_value = [
            {'id': 'fake_id', 'share_network_subnets': [fake_sns]},
        ]
        old = self.ss_active_directory.copy()
        updated = self.ss_active_directory.copy()
        updated['name'] = 'new name'
        updated['description'] = 'new description'
        db.security_service_get.return_value = old
        db.security_service_update.return_value = updated
        body = {
            'security_service': {
                'description': 'new description',
                'name': 'new name',
            },
        }
        req = fakes.HTTPRequest.blank('/security_services/1')
        res_dict = self.controller.update(req, 1, body)['security_service']
        self.assertEqual(updated['description'], res_dict['description'])
        self.assertEqual(updated['name'], res_dict['name'])
        db.security_service_get.assert_called_once_with(
            req.environ['manila.context'], 1)
        db.share_network_get_all_by_security_service.assert_called_once_with(
            req.environ['manila.context'], 1)
        db.security_service_update.assert_called_once_with(
            req.environ['manila.context'], 1, body['security_service'])
        self.assertEqual(2, security_service.policy.check_policy.call_count)
        security_service.policy.check_policy.assert_has_calls([
            mock.call(req.environ['manila.context'],
                      security_service.RESOURCE_NAME, 'update', old)
        ])

    @mock.patch.object(db, 'security_service_get', mock.Mock())
    def test_security_service_update_has_share_servers(self):
        db.security_service_get = mock.Mock()
        self.mock_object(
            self.controller, '_share_servers_dependent_on_sn_exist',
            mock.Mock(return_value=True))
        body = {"security_service": {"type": "ldap"}}

        req = fakes.HTTPRequest.blank('/security_services/1')
        self.assertRaises(webob.exc.HTTPForbidden,
                          self.controller.update,
                          req,
                          1,
                          body)

    @ddt.data(True, False)
    def test_security_service_update_share_server_dependent_exists(self,
                                                                   expected):
        req = fakes.HTTPRequest.blank('/security_services/1')
        context = req.environ['manila.context']
        db.security_service_get = mock.Mock()
        network = (self.fake_share_network_list_with_share_servers if expected
                   else self.fake_share_network_list_without_share_servers)
        db.share_network_get_all_by_security_service = mock.Mock(
            return_value=network)

        result = self.controller._share_servers_dependent_on_sn_exist(
            context, 'fake_id')
        self.assertEqual(expected, result)

    def test_security_service_list(self):
        db.security_service_get_all_by_project = mock.Mock(
            return_value=[self.ss_active_directory.copy()])
        req = fakes.HTTPRequest.blank('/security_services')
        res_dict = self.controller.index(req)
        self.assertEqual(self.security_service_list_expected_resp, res_dict)

    @mock.patch.object(db, 'share_network_get', mock.Mock())
    def test_security_service_list_filter_by_sn(self):
        sn = {
            'id': 'fake_sn_id',
            'security_services': [self.ss_active_directory, ],
        }
        db.share_network_get.return_value = sn
        req = fakes.HTTPRequest.blank(
            '/security-services?share_network_id=fake_sn_id')
        res_dict = self.controller.index(req)
        self.assertEqual(self.security_service_list_expected_resp, res_dict)
        db.share_network_get.assert_called_once_with(
            req.environ['manila.context'],
            sn['id'])

    @mock.patch.object(db, 'security_service_get_all', mock.Mock())
    def test_security_services_list_all_tenants_admin_context(self):
        self.check_policy_patcher.stop()
        db.security_service_get_all.return_value = [
            self.ss_active_directory,
            self.ss_ldap,
        ]
        req = fakes.HTTPRequest.blank(
            '/security-services?all_tenants=1&name=fake-name',
            use_admin_context=True)
        res_dict = self.controller.index(req)
        self.assertEqual(self.security_service_list_expected_resp, res_dict)
        db.security_service_get_all.assert_called_once_with(
            req.environ['manila.context'])

    @mock.patch.object(db, 'security_service_get_all_by_project', mock.Mock())
    def test_security_services_list_all_tenants_non_admin_context(self):
        db.security_service_get_all_by_project.return_value = []
        req = fakes.HTTPRequest.blank(
            '/security-services?all_tenants=1')
        fake_context = req.environ['manila.context']
        self.controller.index(req)
        db.security_service_get_all_by_project.assert_called_once_with(
            fake_context, fake_context.project_id
        )

    @mock.patch.object(db, 'security_service_get_all', mock.Mock())
    def test_security_services_list_all_tenants_with_invalid_value(self):
        req = fakes.HTTPRequest.blank(
            '/security-services?all_tenants=nerd',
            use_admin_context=True)
        self.assertRaises(exception.InvalidInput, self.controller.index, req)

    @mock.patch.object(db, 'security_service_get_all_by_project', mock.Mock())
    def test_security_services_list_all_tenants_with_value_zero(self):
        db.security_service_get_all_by_project.return_value = []
        req = fakes.HTTPRequest.blank(
            '/security-services?all_tenants=0',
            use_admin_context=True)
        res_dict = self.controller.index(req)
        self.assertEqual({'security_services': []}, res_dict)
        db.security_service_get_all_by_project.assert_called_once_with(
            req.environ['manila.context'],
            req.environ['manila.context'].project_id)

    @mock.patch.object(db, 'security_service_get_all_by_project', mock.Mock())
    def test_security_services_list_admin_context_invalid_opts(self):
        db.security_service_get_all_by_project.return_value = [
            self.ss_active_directory,
            self.ss_ldap,
        ]
        req = fakes.HTTPRequest.blank(
            '/security-services?fake_opt=fake_value',
            use_admin_context=True)
        res_dict = self.controller.index(req)
        self.assertEqual({'security_services': []}, res_dict)
        db.security_service_get_all_by_project.assert_called_once_with(
            req.environ['manila.context'],
            req.environ['manila.context'].project_id)

    @mock.patch.object(db, 'security_service_get_all_by_project', mock.Mock())
    def test_security_service_list_all_filter_opts_separately(self):
        db.security_service_get_all_by_project.return_value = [
            self.ss_active_directory,
            self.ss_ldap,
        ]
        for opt, val in self.valid_search_opts.items():
            for use_admin_context in [True, False]:
                req = fakes.HTTPRequest.blank(
                    '/security-services?' + opt + '=' + val,
                    use_admin_context=use_admin_context)
                res_dict = self.controller.index(req)
                self.assertEqual(self.security_service_list_expected_resp,
                                 res_dict)
                db.security_service_get_all_by_project.assert_called_with(
                    req.environ['manila.context'],
                    req.environ['manila.context'].project_id)

    @mock.patch.object(db, 'security_service_get_all_by_project', mock.Mock())
    def test_security_service_list_all_filter_opts(self):
        db.security_service_get_all_by_project.return_value = [
            self.ss_active_directory,
            self.ss_ldap,
        ]
        query_string = '/security-services?' + parse.urlencode(sorted(
            [(k, v) for (k, v) in list(self.valid_search_opts.items())]))
        for use_admin_context in [True, False]:
            req = fakes.HTTPRequest.blank(query_string,
                                          use_admin_context=use_admin_context)
            res_dict = self.controller.index(req)
            self.assertEqual(self.security_service_list_expected_resp,
                             res_dict)
            db.security_service_get_all_by_project.assert_called_with(
                req.environ['manila.context'],
                req.environ['manila.context'].project_id)



# Copyright (c) 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import uuid

from oslo_log import log
from oslo_utils import excutils

from manila import exception
from manila.i18n import _
from manila.share import driver


LOG = log.getLogger(__name__)


class DockerExecHelper(driver.ExecuteMixin):
    def __init__(self, *args, **kwargs):
        self.configuration = kwargs.pop("configuration", None)
        super(DockerExecHelper, self).__init__(*args, **kwargs)
        self.init_execute_mixin()

    def start_container(self, name=None):
        name = name or "".join(["manila_cifs_docker_container",
                                str(uuid.uuid1()).replace("-", "_")])
        image_name = self.configuration.container_image_name
        LOG.debug("Starting container from image %s.", image_name)
        # (aovchinnikov): --privileged is required for both samba and
        # nfs-ganesha to actually allow access to shared folders.
        #
        # (aovchinnikov): To actually make docker container mount a
        # logical volume created after container start-up to some location
        # inside it, we must share entire /dev with it. While seemingly
        # dangerous it is not and moreover this is apparently the only sane
        # way to do it. The reason is when a logical volume gets created
        # several new things appear in /dev: a new /dev/dm-X and a symlink
        # in /dev/volume_group_name pointing to /dev/dm-X. But to be able
        # to interact with /dev/dm-X, it must be already present inside
        # the container's /dev i.e. it must have been -v shared during
        # container start-up. So we should either precreate an unknown
        # number of /dev/dm-Xs (one per LV), share them all and hope
        # for the best or share the entire /dev and hope for the best.
        #
        # The risk of allowing a container having access to entire host's
        # /dev is not as big as it seems: as long as actual share providers
        # are invulnerable this does not pose any extra risks. If, however,
        # share providers contain vulnerabilities then the driver does not
        # provide any more possibilities for an exploitation than other
        # first-party drivers.

        path = "{0}:/shares".format(
            self.configuration.container_volume_mount_path)
        cmd = ["docker", "run", "-d", "-i", "-t", "--privileged",
               "-v", "/dev:/dev", "--name=%s" % name,
               "-v", path, image_name]
        try:
            result = self._inner_execute(cmd)
        except (exception.ProcessExecutionError, OSError):
            raise exception.ShareBackendException(
                msg="Container %s has failed to start." % name)
        LOG.info("A container has been successfully started! Its id is "
                 "%s.", result[0].rstrip('\n'))

    def stop_container(self, name):
        LOG.debug("Stopping container %s.", name)
        try:
            self._inner_execute(["docker", "stop", name])
        except (exception.ProcessExecutionError, OSError):
            raise exception.ShareBackendException(
                msg="Container %s has failed to stop properly." % name)
        LOG.info("Container %s is successfully stopped.", name)

    def execute(self, name=None, cmd=None, ignore_errors=False):
        if name is None:
            raise exception.ManilaException(_("Container name not specified."))
        if cmd is None or (type(cmd) is not list):
            raise exception.ManilaException(_("Missing or malformed command."))
        LOG.debug("Executing inside a container %s.", name)
        cmd = ["docker", "exec", "-i", name] + cmd
        result = self._inner_execute(cmd, ignore_errors=ignore_errors)
        return result

    def _inner_execute(self, cmd, ignore_errors=False):
        LOG.debug("Executing command: %s.", " ".join(cmd))
        try:
            result = self._execute(*cmd, run_as_root=True)
        except (exception.ProcessExecutionError, OSError) as e:
            with excutils.save_and_reraise_exception(
                    reraise=not ignore_errors):
                LOG.warning("Failed to run command %(cmd)s due to "
                            "%(reason)s.", {'cmd': cmd, 'reason': e})
        else:
            LOG.debug("Execution result: %s.", result)
            return result

    def fetch_container_address(self, name, address_family="inet6"):
        result = self.execute(
            name,
            ["ip", "-oneline",
             "-family", address_family,
             "address", "show", "scope", "global", "dev", "eth0"],
        )
        address_w_prefix = result[0].split()[3]
        address = address_w_prefix.split('/')[0]
        return address

    def rename_container(self, name, new_name):
        veth_name = self.find_container_veth(name)
        if not veth_name:
            raise exception.ManilaException(
                _("Could not find OVS information related to "
                  "container %s.") % name)

        try:
            self._inner_execute(["docker", "rename", name, new_name])
        except (exception.ProcessExecutionError, OSError):
            raise exception.ShareBackendException(
                msg="Could not rename container %s." % name)

        cmd = ["ovs-vsctl", "set", "interface", veth_name,
               "external-ids:manila-container=%s" % new_name]
        try:
            self._inner_execute(cmd)
        except (exception.ProcessExecutionError, OSError):
            try:
                self._inner_execute(["docker", "rename", new_name, name])
            except (exception.ProcessExecutionError, OSError):
                msg = _("Could not rename back container %s.") % name
                LOG.exception(msg)
            raise exception.ShareBackendException(
                msg="Could not update OVS information %s." % name)

        LOG.info("Container %s has been successfully renamed.", name)

    def find_container_veth(self, name):
        interfaces = self._execute("ovs-vsctl", "list", "interface",
                                   run_as_root=True)[0]
        veths = set(re.findall("veth[0-9a-zA-Z]{7}", interfaces))
        manila_re = "manila-container=\"?.{%s}\"?" % len(name)
        for veth in veths:
            try:
                iface_data = self._execute("ovs-vsctl", "list", "interface",
                                           veth, run_as_root=True)[0]
            except (exception.ProcessExecutionError, OSError) as e:
                LOG.debug("Error listing interface %(veth)s. "
                          "Reason: %(reason)s", {'veth': veth,
                                                 'reason': e})
                continue

            container_id = re.findall(manila_re, iface_data)
            if container_id == []:
                continue
            elif container_id[0].split("manila-container=")[-1].split(
                    "manila_")[-1].strip('"') == name.split("manila_")[-1]:
                return veth

    def container_exists(self, name):

        result = self._execute("docker", "ps", "--no-trunc",
                               "--format='{{.Names}}'", run_as_root=True)[0]
        for line in result.split('\n'):
            if name == line.strip("'"):
                return True
        return False



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from lxml import etree
import six


class XMLAPIParser(object):
    def __init__(self):
        # The following Boolean acts as the flag for the common sub-element.
        # For instance:
        #     <CifsServers>
        #         <li> server_1 </li>
        #     </CifsServers>
        #     <Alias>
        #         <li> interface_1 </li>
        #     </Alias>
        self.is_QueryStatus = False
        self.is_CifsServers = False
        self.is_Aliases = False
        self.is_MoverStatus = False
        self.is_TaskResponse = False
        self.is_Vdm = False
        self.is_Interfaces = False

        self.elt = {}

    def _remove_ns(self, tag):
        i = tag.find('}')
        if i >= 0:
            tag = tag[i + 1:]
        return tag

    def parse(self, xml):
        result = {
            'type': None,
            'taskId': None,
            'maxSeverity': None,
            'objects': [],
            'problems': [],
        }

        events = ("start", "end")

        context = etree.iterparse(six.BytesIO(xml),
                                  events=events)
        for action, elem in context:
            self.tag = self._remove_ns(elem.tag)

            func = self._get_func(action, self.tag)
            if func in vars(XMLAPIParser):
                if action == 'start':
                    eval('self.' + func)(elem, result)
                elif action == 'end':
                    eval('self.' + func)()

        return result

    def _get_func(self, action, tag):
        if tag == 'W2KServerData':
            return action + '_' + 'w2k_server_data'

        temp_list = re.sub(r"([A-Z])", r" \1", tag).split()
        if temp_list:
            func_name = action + '_' + '_'.join(temp_list)
        else:
            func_name = action + '_' + tag
        return func_name.lower()

    def _copy_property(self, source, target, property, list_property=None):
        for key in property:
            if key in source:
                target[key] = source[key]

        if list_property:
            for key in list_property:
                if key in source:
                    target[key] = source[key].split()

    def _append_elm_property(self, elm, result, property, identifier):
        for obj in result['objects']:
            if (identifier in obj and identifier in elm.attrib and
                    elm.attrib[identifier] == obj[identifier]):
                for key, value in elm.attrib.items():
                    if key in property:
                        obj[key] = value

    def _append_element(self, elm, result, property, list_property,
                        identifier):
        sub_elm = {}
        self._copy_property(elm.attrib, sub_elm, property, list_property)

        for obj in result['objects']:
            if (identifier in obj and identifier in elm.attrib and
                    elm.attrib[identifier] == obj[identifier]):
                if self.tag in obj:
                    obj[self.tag].append(sub_elm)
                else:
                    obj[self.tag] = [sub_elm]

    def start_task_response(self, elm, result):
        self.is_TaskResponse = True
        result['type'] = 'TaskResponse'
        self._copy_property(elm.attrib, result, ['taskId'])

    def end_task_response(self):
        self.is_TaskResponse = False

    def start_fault(self, elm, result):
        result['type'] = 'Fault'

    def start_status(self, elm, result):
        if self.is_TaskResponse:
            result['maxSeverity'] = elm.attrib['maxSeverity']
        elif self.is_MoverStatus or self.is_Vdm:
            self.elt['maxSeverity'] = elm.attrib['maxSeverity']

    def start_query_status(self, elm, result):
        self.is_QueryStatus = True
        result['type'] = 'QueryStatus'
        self._copy_property(elm.attrib, result, ['maxSeverity'])

    def end_query_status(self):
        self.is_QueryStatus = False

    def start_problem(self, elm, result):
        self.elt = {}
        properties = ('message', 'messageCode')

        self._copy_property(elm.attrib, self.elt, properties)
        result['problems'].append(self.elt)

    def start_description(self, elm, result):
        self.elt['Description'] = elm.text

    def start_action(self, elm, result):
        self.elt['Action'] = elm.text

    def start_diagnostics(self, elm, result):
        self.elt['Diagnostics'] = elm.text

    def start_file_system(self, elm, result):
        self.elt = {}
        property = (
            'fileSystem',
            'name',
            'type',
            'storages',
            'volume',
            'dataServicePolicies',
            'internalUse',
        )
        list_property = ('storagePools',)

        self._copy_property(elm.attrib, self.elt, property, list_property)
        result['objects'].append(self.elt)

    def start_file_system_capacity_info(self, elm, result):
        property = ('volumeSize',)

        identifier = 'fileSystem'

        self._append_elm_property(elm, result, property, identifier)

    def start_storage_pool(self, elm, result):
        self.elt = {}
        property = ('name', 'autoSize', 'usedSize', 'diskType', 'pool',
                    'dataServicePolicies', 'virtualProvisioning')
        list_property = ('movers',)

        self._copy_property(elm.attrib, self.elt, property, list_property)
        result['objects'].append(self.elt)

    def start_system_storage_pool_data(self, elm, result):
        property = ('greedy', 'isBackendPool')

        self._copy_property(elm.attrib, self.elt, property)

    def start_mover(self, elm, result):
        self.elt = {}
        property = ('name', 'host', 'mover', 'role')
        list_property = ('ntpServers', 'standbyFors', 'standbys')

        self._copy_property(elm.attrib, self.elt, property, list_property)
        result['objects'].append(self.elt)

    def start_mover_status(self, elm, result):
        self.is_MoverStatus = True

        property = ('version', 'csTime', 'clock', 'timezone', 'uptime')

        identifier = 'mover'

        self._append_elm_property(elm, result, property, identifier)

    def end_mover_status(self):
        self.is_MoverStatus = False

    def start_mover_dns_domain(self, elm, result):
        property = ('name', 'protocol')
        list_property = ('servers',)

        identifier = 'mover'

        self._append_element(elm, result, property, list_property, identifier)

    def start_mover_interface(self, elm, result):
        property = (
            'name',
            'device',
            'up',
            'ipVersion',
            'netMask',
            'ipAddress',
            'vlanid',
        )

        identifier = 'mover'

        self._append_element(elm, result, property, None, identifier)

    def start_logical_network_device(self, elm, result):
        property = ('name', 'type', 'speed')
        list_property = ('interfaces',)
        identifier = 'mover'

        self._append_element(elm, result, property, list_property, identifier)

    def start_vdm(self, elm, result):
        self.is_Vdm = True

        self.elt = {}
        property = ('name', 'state', 'mover', 'vdm')

        self._copy_property(elm.attrib, self.elt, property)
        result['objects'].append(self.elt)

    def end_vdm(self):
        self.is_Vdm = False

    def start_interfaces(self, elm, result):
        self.is_Interfaces = True
        self.elt['Interfaces'] = []

    def end_interfaces(self):
        self.is_Interfaces = False

    def start_li(self, elm, result):
        if self.is_CifsServers:
            self.elt['CifsServers'].append(elm.text)
        elif self.is_Aliases:
            self.elt['Aliases'].append(elm.text)
        elif self.is_Interfaces:
            self.elt['Interfaces'].append(elm.text)

    def start_cifs_server(self, elm, result):
        self.elt = {}
        property = ('type', 'localUsers', 'name', 'mover', 'moverIdIsVdm')

        list_property = ('interfaces',)

        self._copy_property(elm.attrib, self.elt, property, list_property)
        result['objects'].append(self.elt)

    def start_aliases(self, elm, result):
        self.is_Aliases = True
        self.elt['Aliases'] = []

    def end_aliases(self):
        self.is_Aliases = False

    def start_w2k_server_data(self, elm, result):
        property = ('domain', 'compName', 'domainJoined')

        self._copy_property(elm.attrib, self.elt, property)

    def start_cifs_share(self, elm, result):
        self.elt = {}
        property = ('path', 'fileSystem', 'name', 'mover', 'moverIdIsVdm')

        self._copy_property(elm.attrib, self.elt, property)
        result['objects'].append(self.elt)

    def start_cifs_servers(self, elm, result):
        self.is_CifsServers = True
        self.elt['CifsServers'] = []

    def end_cifs_servers(self):
        self.is_CifsServers = False

    def start_checkpoint(self, elm, result):
        self.elt = {}
        property = ('checkpointOf', 'name', 'checkpoint', 'state')

        self._copy_property(elm.attrib, self.elt, property)
        result['objects'].append(self.elt)

    def start_mount(self, elm, result):
        self.elt = {}
        property = ('fileSystem', 'path', 'mover', 'moverIdIsVdm')

        self._copy_property(elm.attrib, self.elt, property)
        result['objects'].append(self.elt)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
from six.moves import http_client
import webob
from webob import exc

from manila.api.openstack import wsgi
from manila.api.views import share_servers as share_servers_views
from manila.common import constants
from manila.db import api as db_api
from manila import exception
from manila.i18n import _
from manila import share

LOG = log.getLogger(__name__)


class ShareServerController(wsgi.Controller):
    """The Share Server API controller for the OpenStack API."""

    _view_builder_class = share_servers_views.ViewBuilder
    resource_name = 'share_server'

    def __init__(self):
        self.share_api = share.API()
        super(ShareServerController, self).__init__()

    @wsgi.Controller.authorize
    def index(self, req):
        """Returns a list of share servers."""

        context = req.environ['manila.context']

        search_opts = {}
        search_opts.update(req.GET)
        share_servers = db_api.share_server_get_all(context)
        for s in share_servers:
            try:
                s.share_network_id = s.share_network_subnet['share_network_id']
                share_network = db_api.share_network_get(
                    context, s.share_network_id)
                s.project_id = share_network['project_id']
                if share_network['name']:
                    s.share_network_name = share_network['name']
                else:
                    s.share_network_name = share_network['id']
            except exception.ShareNetworkNotFound:
                # NOTE(dviroel): The share-network may already be deleted while
                # the share-server is in 'deleting' state. In this scenario,
                # we will return some empty values.
                LOG.debug("Unable to retrieve share network details for share "
                          "server %(server)s, the network %(network)s was "
                          "not found.",
                          {'server': s.id, 'network': s.share_network_id})
                s.project_id = ''
                s.share_network_name = ''
        if search_opts:
            for k, v in search_opts.items():
                share_servers = [s for s in share_servers if
                                 (hasattr(s, k) and
                                  s[k] == v or k == 'share_network' and
                                  v in [s.share_network_name,
                                        s.share_network_id])]
        return self._view_builder.build_share_servers(req, share_servers)

    @wsgi.Controller.authorize
    def show(self, req, id):
        """Return data about the requested share server."""
        context = req.environ['manila.context']
        try:
            server = db_api.share_server_get(context, id)
            share_network = db_api.share_network_get(
                context, server.share_network_subnet['share_network_id'])
            server.share_network_id = share_network['id']
            server.project_id = share_network['project_id']
            if share_network['name']:
                server.share_network_name = share_network['name']
            else:
                server.share_network_name = share_network['id']
        except exception.ShareServerNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)
        except exception.ShareNetworkNotFound:
            msg = _("Share server %s could not be found. Its associated "
                    "share network does not "
                    "exist.") % server.share_network_subnet['share_network_id']
            raise exc.HTTPNotFound(explanation=msg)
        return self._view_builder.build_share_server(req, server)

    @wsgi.Controller.authorize
    def details(self, req, id):
        """Return details for requested share server."""
        context = req.environ['manila.context']
        try:
            share_server = db_api.share_server_get(context, id)
        except exception.ShareServerNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        return self._view_builder.build_share_server_details(
            share_server['backend_details'])

    @wsgi.Controller.authorize
    def delete(self, req, id):
        """Delete specified share server."""
        context = req.environ['manila.context']
        try:
            share_server = db_api.share_server_get(context, id)
        except exception.ShareServerNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)
        allowed_statuses = [constants.STATUS_ERROR, constants.STATUS_ACTIVE]
        if share_server['status'] not in allowed_statuses:
            data = {
                'status': share_server['status'],
                'allowed_statuses': allowed_statuses,
            }
            msg = _("Share server's actual status is %(status)s, allowed "
                    "statuses for deletion are %(allowed_statuses)s.") % (data)
            raise exc.HTTPForbidden(explanation=msg)
        LOG.debug("Deleting share server with id: %s.", id)
        try:
            self.share_api.delete_share_server(context, share_server)
        except exception.ShareServerInUse as e:
            raise exc.HTTPConflict(explanation=e.msg)
        return webob.Response(status_int=http_client.ACCEPTED)


def create_resource():
    return wsgi.Resource(ShareServerController())



# Copyright (c) 2014 Navneet Singh.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
NetApp API for Data ONTAP and OnCommand DFM.

Contains classes required to issue API calls to Data ONTAP and OnCommand DFM.
"""

import copy
import re

from lxml import etree
from oslo_log import log
from oslo_serialization import jsonutils
import requests
from requests import auth
import six

from manila import exception
from manila.i18n import _
from manila.share.drivers.netapp.dataontap.client import rest_endpoints
from manila.share.drivers.netapp import utils

LOG = log.getLogger(__name__)

EONTAPI_EINVAL = '22'
EVOLOPNOTSUPP = '160'
EAPIERROR = '13001'
EAPINOTFOUND = '13005'
ESNAPSHOTNOTALLOWED = '13023'
EVOLUMEOFFLINE = '13042'
EINTERNALERROR = '13114'
EINVALIDINPUTERROR = '13115'
EDUPLICATEENTRY = '13130'
EVOLNOTCLONE = '13170'
EVOLMOVE_CANNOT_MOVE_TO_CFO = '13633'
EAGGRDOESNOTEXIST = '14420'
EVOL_NOT_MOUNTED = '14716'
EVSERVERALREADYSTARTED = '14923'
ESIS_CLONE_NOT_LICENSED = '14956'
EOBJECTNOTFOUND = '15661'
EVSERVERNOTFOUND = '15698'
E_VIFMGR_PORT_ALREADY_ASSIGNED_TO_BROADCAST_DOMAIN = '18605'
ERELATION_EXISTS = '17122'
ENOTRANSFER_IN_PROGRESS = '17130'
ETRANSFER_IN_PROGRESS = '17137'
EANOTHER_OP_ACTIVE = '17131'
ERELATION_NOT_QUIESCED = '17127'
ESOURCE_IS_DIFFERENT = '17105'
EVOL_CLONE_BEING_SPLIT = '17151'
EPOLICYNOTFOUND = '18251'
EEVENTNOTFOUND = '18253'
ESCOPENOTFOUND = '18259'
ESVMDR_CANNOT_PERFORM_OP_FOR_STATUS = '18815'
ENFS_V4_0_ENABLED_MIGRATION_FAILURE = '13172940'
EVSERVER_MIGRATION_TO_NON_AFF_CLUSTER = '13172984'

STYLE_LOGIN_PASSWORD = 'basic_auth'
TRANSPORT_TYPE_HTTP = 'http'
TRANSPORT_TYPE_HTTPS = 'https'
STYLE_CERTIFICATE = 'certificate_auth'


class BaseClient(object):
    """Encapsulates server connection logic."""

    def __init__(self, host, transport_type=TRANSPORT_TYPE_HTTP, style=None,
                 ssl_cert_path=None, username=None, password=None, port=None,
                 trace=False, api_trace_pattern=None):
        super(BaseClient, self).__init__()
        self._host = host
        self.set_transport_type(transport_type)
        self.set_style(style)
        if port:
            self.set_port(port)
        self._username = username
        self._password = password
        self._trace = trace
        self._api_trace_pattern = api_trace_pattern
        self._refresh_conn = True
        if ssl_cert_path is not None:
            self._ssl_verify = ssl_cert_path
        else:
            # Note(felipe_rodrigues): it will verify with the mozila CA roots,
            # given by certifi package.
            self._ssl_verify = True
        LOG.debug('Using NetApp controller: %s', self._host)

    def get_style(self):
        """Get the authorization style for communicating with the server."""
        return self._auth_style

    def set_style(self, style):
        """Set the authorization style for communicating with the server.

        Supports basic_auth for now. Certificate_auth mode to be done.
        """
        if style.lower() not in (STYLE_LOGIN_PASSWORD, STYLE_CERTIFICATE):
            raise ValueError('Unsupported authentication style')
        self._auth_style = style.lower()

    def get_transport_type(self):
        """Get the transport type protocol."""
        return self._protocol

    def set_transport_type(self, transport_type):
        """Set the transport type protocol for API.

        Supports http and https transport types.
        """
        if transport_type.lower() not in (
                TRANSPORT_TYPE_HTTP, TRANSPORT_TYPE_HTTPS):
            raise ValueError('Unsupported transport type')
        self._protocol = transport_type.lower()
        self._refresh_conn = True

    def get_server_type(self):
        """Get the server type."""
        return self._server_type

    def set_server_type(self, server_type):
        """Set the target server type.

        Supports filer and dfm server types.
        """
        raise NotImplementedError()

    def set_api_version(self, major, minor):
        """Set the API version."""
        try:
            self._api_major_version = int(major)
            self._api_minor_version = int(minor)
            self._api_version = (six.text_type(major) + "." +
                                 six.text_type(minor))
        except ValueError:
            raise ValueError('Major and minor versions must be integers')
        self._refresh_conn = True

    def set_system_version(self, system_version):
        """Set the ONTAP system version."""
        self._system_version = system_version
        self._refresh_conn = True

    def get_api_version(self):
        """Gets the API version tuple."""
        if hasattr(self, '_api_version'):
            return (self._api_major_version, self._api_minor_version)
        return None

    def get_system_version(self):
        """Gets the ONTAP system version."""
        if hasattr(self, '_system_version'):
            return self._system_version
        return None

    def set_port(self, port):
        """Set the server communication port."""
        try:
            int(port)
        except ValueError:
            raise ValueError('Port must be integer')
        self._port = six.text_type(port)
        self._refresh_conn = True

    def get_port(self):
        """Get the server communication port."""
        return self._port

    def set_timeout(self, seconds):
        """Sets the timeout in seconds."""
        try:
            self._timeout = int(seconds)
        except ValueError:
            raise ValueError('timeout in seconds must be integer')

    def get_timeout(self):
        """Gets the timeout in seconds if set."""
        if hasattr(self, '_timeout'):
            return self._timeout
        return None

    def get_vserver(self):
        """Get the vserver to use in tunneling."""
        return self._vserver

    def set_vserver(self, vserver):
        """Set the vserver to use if tunneling gets enabled."""
        self._vserver = vserver

    def set_username(self, username):
        """Set the user name for authentication."""
        self._username = username
        self._refresh_conn = True

    def set_password(self, password):
        """Set the password for authentication."""
        self._password = password
        self._refresh_conn = True

    def invoke_successfully(self, na_element, api_args=None,
                            enable_tunneling=False, use_zapi=True):
        """Invokes API and checks execution status as success.

        Need to set enable_tunneling to True explicitly to achieve it.
        This helps to use same connection instance to enable or disable
        tunneling. The vserver or vfiler should be set before this call
        otherwise tunneling remains disabled.
        """
        pass

    def _build_session(self):
        """Builds a session in the client."""
        if self._auth_style == STYLE_LOGIN_PASSWORD:
            auth_handler = self._create_basic_auth_handler()
        else:
            auth_handler = self._create_certificate_auth_handler()

        self._session = requests.Session()
        self._session.auth = auth_handler
        self._session.verify = self._ssl_verify
        headers = self._build_headers()

        self._session.headers = headers

    def _build_headers(self):
        """Adds the necessary headers to the session."""
        raise NotImplementedError()

    def _create_basic_auth_handler(self):
        """Creates and returns a basic HTTP auth handler."""
        return auth.HTTPBasicAuth(self._username, self._password)

    def _create_certificate_auth_handler(self):
        """Creates and returns a certificate auth handler."""
        raise NotImplementedError()

    def __str__(self):
        """Gets a representation of the client."""
        return "server: %s" % (self._host)


class ZapiClient(BaseClient):

    SERVER_TYPE_FILER = 'filer'
    SERVER_TYPE_DFM = 'dfm'
    URL_FILER = 'servlets/netapp.servlets.admin.XMLrequest_filer'
    URL_DFM = 'apis/XMLrequest'
    NETAPP_NS = 'http://www.netapp.com/filer/admin'

    def __init__(self, host, server_type=SERVER_TYPE_FILER,
                 transport_type=TRANSPORT_TYPE_HTTP,
                 style=STYLE_LOGIN_PASSWORD, ssl_cert_path=None, username=None,
                 password=None, port=None, trace=False,
                 api_trace_pattern=utils.API_TRACE_PATTERN):
        super(ZapiClient, self).__init__(
            host, transport_type=transport_type, style=style,
            ssl_cert_path=ssl_cert_path, username=username, password=password,
            port=port, trace=trace, api_trace_pattern=api_trace_pattern)
        self.set_server_type(server_type)
        if port is None:
            # Not yet set in parent, use defaults
            self._set_port()

    def _set_port(self):
        """Defines which port will be used to communicate with ONTAP."""
        if self._protocol == TRANSPORT_TYPE_HTTP:
            if self._server_type == ZapiClient.SERVER_TYPE_FILER:
                self.set_port(80)
            else:
                self.set_port(8088)
        else:
            if self._server_type == ZapiClient.SERVER_TYPE_FILER:
                self.set_port(443)
            else:
                self.set_port(8488)

    def set_server_type(self, server_type):
        """Set the target server type.

        Supports filer and dfm server types.
        """
        if server_type.lower() not in (ZapiClient.SERVER_TYPE_FILER,
                                       ZapiClient.SERVER_TYPE_DFM):
            raise ValueError('Unsupported server type')
        self._server_type = server_type.lower()
        if self._server_type == ZapiClient.SERVER_TYPE_FILER:
            self._url = ZapiClient.URL_FILER
        else:
            self._url = ZapiClient.URL_DFM
        self._ns = ZapiClient.NETAPP_NS
        self._refresh_conn = True

    def get_vfiler(self):
        """Get the vfiler to use in tunneling."""
        return self._vfiler

    def set_vfiler(self, vfiler):
        """Set the vfiler to use if tunneling gets enabled."""
        self._vfiler = vfiler

    def invoke_elem(self, na_element, enable_tunneling=False):
        """Invoke the API on the server."""
        if na_element and not isinstance(na_element, NaElement):
            ValueError('NaElement must be supplied to invoke API')

        request_element = self._create_request(na_element, enable_tunneling)
        request_d = request_element.to_string()

        api_name = na_element.get_name()
        api_name_matches_regex = (re.match(self._api_trace_pattern, api_name)
                                  is not None)

        if self._trace and api_name_matches_regex:
            LOG.debug("Request: %s", request_element.to_string(pretty=True))

        if (not hasattr(self, '_session') or not self._session
                or self._refresh_conn):
            self._build_session()
        try:
            if hasattr(self, '_timeout'):
                response = self._session.post(
                    self._get_url(), data=request_d, timeout=self._timeout)
            else:
                response = self._session.post(
                    self._get_url(), data=request_d)
        except requests.HTTPError as e:
            raise NaApiError(e.errno, e.strerror)
        except requests.URLRequired as e:
            raise exception.StorageCommunicationException(six.text_type(e))
        except Exception as e:
            raise NaApiError(message=e)

        response_xml = response.text
        response_element = self._get_result(
            bytes(bytearray(response_xml, encoding='utf-8')))

        if self._trace and api_name_matches_regex:
            LOG.debug("Response: %s", response_element.to_string(pretty=True))

        return response_element

    def invoke_successfully(self, na_element, api_args=None,
                            enable_tunneling=False, use_zapi=True):
        """Invokes API and checks execution status as success.

        Need to set enable_tunneling to True explicitly to achieve it.
        This helps to use same connection instance to enable or disable
        tunneling. The vserver or vfiler should be set before this call
        otherwise tunneling remains disabled.
        """
        if api_args:
            na_element.translate_struct(api_args)

        result = self.invoke_elem(
            na_element, enable_tunneling=enable_tunneling)

        if result.has_attr('status') and result.get_attr('status') == 'passed':
            return result
        code = (result.get_attr('errno')
                or result.get_child_content('errorno')
                or 'ESTATUSFAILED')
        if code == ESIS_CLONE_NOT_LICENSED:
            msg = 'Clone operation failed: FlexClone not licensed.'
        else:
            msg = (result.get_attr('reason')
                   or result.get_child_content('reason')
                   or 'Execution status is failed due to unknown reason')
        raise NaApiError(code, msg)

    def _create_request(self, na_element, enable_tunneling=False):
        """Creates request in the desired format."""
        netapp_elem = NaElement('netapp')
        netapp_elem.add_attr('xmlns', self._ns)
        if hasattr(self, '_api_version'):
            netapp_elem.add_attr('version', self._api_version)
        if enable_tunneling:
            self._enable_tunnel_request(netapp_elem)
        netapp_elem.add_child_elem(na_element)
        return netapp_elem

    def _enable_tunnel_request(self, netapp_elem):
        """Enables vserver or vfiler tunneling."""
        if hasattr(self, '_vfiler') and self._vfiler:
            if (hasattr(self, '_api_major_version') and
                    hasattr(self, '_api_minor_version') and
                    self._api_major_version >= 1 and
                    self._api_minor_version >= 7):
                netapp_elem.add_attr('vfiler', self._vfiler)
            else:
                raise ValueError('ontapi version has to be atleast 1.7'
                                 ' to send request to vfiler')
        if hasattr(self, '_vserver') and self._vserver:
            if (hasattr(self, '_api_major_version') and
                    hasattr(self, '_api_minor_version') and
                    self._api_major_version >= 1 and
                    self._api_minor_version >= 15):
                netapp_elem.add_attr('vfiler', self._vserver)
            else:
                raise ValueError('ontapi version has to be atleast 1.15'
                                 ' to send request to vserver')

    @staticmethod
    def _parse_response(response):
        """Get the NaElement for the response."""
        if not response:
            raise NaApiError('No response received')
        xml = etree.XML(response)
        return NaElement(xml)

    def _get_result(self, response):
        """Gets the call result."""
        processed_response = self._parse_response(response)
        return processed_response.get_child_by_name('results')

    def _get_url(self):
        """Get the base url to send the request."""
        host = self._host
        if ':' in host:
            host = '[%s]' % host
        return '%s://%s:%s/%s' % (self._protocol, host, self._port, self._url)

    def _build_headers(self):
        """Build and return headers."""
        return {'Content-Type': 'text/xml'}


class RestClient(BaseClient):

    def __init__(self, host, transport_type=TRANSPORT_TYPE_HTTP,
                 style=STYLE_LOGIN_PASSWORD, ssl_cert_path=None, username=None,
                 password=None, port=None, trace=False,
                 api_trace_pattern=utils.API_TRACE_PATTERN):
        super(RestClient, self).__init__(
            host, transport_type=transport_type, style=style,
            ssl_cert_path=ssl_cert_path, username=username, password=password,
            port=port, trace=trace, api_trace_pattern=api_trace_pattern)
        if port is None:
            # Not yet set in parent, use defaults
            self._set_port()

    def _set_port(self):
        if self._protocol == TRANSPORT_TYPE_HTTP:
            self.set_port(80)
        else:
            self.set_port(443)

    def _get_request_info(self, api_name, session):
        """Returns the request method and url to be used in the REST call."""

        request_methods = {
            'post': session.post,
            'get': session.get,
            'put': session.put,
            'delete': session.delete,
            'patch': session.patch,
        }
        rest_call = rest_endpoints.endpoints.get(api_name)
        return request_methods[rest_call['method']], rest_call['url']

    def _add_query_params_to_url(self, url, query):
        """Populates the URL with specified filters."""
        filters = ""
        for k, v in query.items():
            filters += "%(key)s=%(value)s&" % {"key": k, "value": v}
        url += "?" + filters
        return url

    def invoke_elem(self, na_element, api_args=None):
        """Invoke the API on the server."""
        if na_element and not isinstance(na_element, NaElement):
            raise ValueError('NaElement must be supplied to invoke API')

        api_name = na_element.get_name()
        api_name_matches_regex = (re.match(self._api_trace_pattern, api_name)
                                  is not None)
        data = api_args.get("body") if api_args else {}

        if (not hasattr(self, '_session') or not self._session
                or self._refresh_conn):
            self._build_session()
        request_method, action_url = self._get_request_info(
            api_name, self._session)

        url_params = api_args.get("url_params") if api_args else None
        if url_params:
            action_url = action_url % url_params

        query = api_args.get("query") if api_args else None
        if query:
            action_url = self._add_query_params_to_url(
                action_url, api_args['query'])

        url = self._get_base_url() + action_url
        data = jsonutils.dumps(data) if data else data

        if self._trace and api_name_matches_regex:
            message = ("Request: %(method)s %(url)s. Request body "
                       "%(body)s") % {
                "method": request_method,
                "url": action_url,
                "body": api_args.get("body") if api_args else {}
            }
            LOG.debug(message)

        try:
            if hasattr(self, '_timeout'):
                response = request_method(
                    url, data=data, timeout=self._timeout)
            else:
                response = request_method(url, data=data)
        except requests.HTTPError as e:
            raise NaApiError(e.errno, e.strerror)
        except requests.URLRequired as e:
            raise exception.StorageCommunicationException(six.text_type(e))
        except Exception as e:
            raise NaApiError(message=e)

        response = (
            jsonutils.loads(response.content) if response.content else None)
        if self._trace and api_name_matches_regex:
            LOG.debug("Response: %s", response)

        return response

    def invoke_successfully(self, na_element, api_args=None,
                            enable_tunneling=False, use_zapi=False):
        """Invokes API and checks execution status as success.

        Need to set enable_tunneling to True explicitly to achieve it.
        This helps to use same connection instance to enable or disable
        tunneling. The vserver or vfiler should be set before this call
        otherwise tunneling remains disabled.
        """
        result = self.invoke_elem(na_element, api_args=api_args)
        if not result.get('error'):
            return result
        result_error = result.get('error')
        code = (result_error.get('code')
                or 'ESTATUSFAILED')
        if code == ESIS_CLONE_NOT_LICENSED:
            msg = 'Clone operation failed: FlexClone not licensed.'
        else:
            msg = (result_error.get('message')
                   or 'Execution status is failed due to unknown reason')
        raise NaApiError(code, msg)

    def _get_base_url(self):
        """Get the base URL for REST requests."""
        host = self._host
        if ':' in host:
            host = '[%s]' % host
        return '%s://%s:%s/api/' % (self._protocol, host, self._port)

    def _build_headers(self):
        """Build and return headers for a REST request."""
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        return headers


class NaServer(object):
    """Encapsulates server connection logic."""

    def __init__(self, host, transport_type=TRANSPORT_TYPE_HTTP,
                 style=STYLE_LOGIN_PASSWORD, ssl_cert_path=None, username=None,
                 password=None, port=None, trace=False,
                 api_trace_pattern=utils.API_TRACE_PATTERN):
        self.zapi_client = ZapiClient(
            host, transport_type=transport_type, style=style,
            ssl_cert_path=ssl_cert_path, username=username, password=password,
            port=port, trace=trace, api_trace_pattern=api_trace_pattern)
        self.rest_client = RestClient(
            host, transport_type=transport_type, style=style,
            ssl_cert_path=ssl_cert_path, username=username, password=password,
            port=port, trace=trace, api_trace_pattern=api_trace_pattern
        )
        self._host = host

        LOG.debug('Using NetApp controller: %s', self._host)

    def get_transport_type(self, use_zapi_client=True):
        """Get the transport type protocol."""
        return self.get_client(use_zapi=use_zapi_client).get_transport_type()

    def set_transport_type(self, transport_type):
        """Set the transport type protocol for API.

        Supports http and https transport types.
        """
        self.zapi_client.set_transport_type(transport_type)
        self.rest_client.set_transport_type(transport_type)

    def get_style(self, use_zapi_client=True):
        """Get the authorization style for communicating with the server."""
        return self.get_client(use_zapi=use_zapi_client).get_style()

    def set_style(self, style):
        """Set the authorization style for communicating with the server.

        Supports basic_auth for now. Certificate_auth mode to be done.
        """
        self.zapi_client.set_style(style)
        self.rest_client.set_style(style)

    def get_server_type(self, use_zapi_client=True):
        """Get the target server type."""
        return self.get_client(use_zapi=use_zapi_client).get_server_type()

    def set_server_type(self, server_type):
        """Set the target server type.

        Supports filer and dfm server types.
        """
        self.zapi_client.set_server_type(server_type)
        self.rest_client.set_server_type(server_type)

    def set_api_version(self, major, minor):
        """Set the API version."""
        self.zapi_client.set_api_version(major, minor)
        self.rest_client.set_api_version(1, 0)

    def set_system_version(self, system_version):
        """Set the ONTAP system version."""
        self.zapi_client.set_system_version(system_version)
        self.rest_client.set_system_version(system_version)

    def get_api_version(self, use_zapi_client=True):
        """Gets the API version tuple."""
        return self.get_client(use_zapi=use_zapi_client).get_api_version()

    def get_system_version(self, use_zapi_client=True):
        """Gets the ONTAP system version."""
        return self.get_client(use_zapi=use_zapi_client).get_system_version()

    def set_port(self, port):
        """Set the server communication port."""
        self.zapi_client.set_port(port)
        self.rest_client.set_port(port)

    def get_port(self, use_zapi_client=True):
        """Get the server communication port."""
        return self.get_client(use_zapi=use_zapi_client).get_port()

    def set_timeout(self, seconds):
        """Sets the timeout in seconds."""
        self.zapi_client.set_timeout(seconds)
        self.rest_client.set_timeout(seconds)

    def get_timeout(self, use_zapi_client=True):
        """Gets the timeout in seconds if set."""
        return self.get_client(use_zapi=use_zapi_client).get_timeout()

    def get_vfiler(self):
        """Get the vfiler to use in tunneling."""
        return self.zapi_client.get_vfiler()

    def set_vfiler(self, vfiler):
        """Set the vfiler to use if tunneling gets enabled."""
        self.zapi_client.set_vfiler(vfiler)

    def get_vserver(self, use_zapi_client=True):
        """Get the vserver to use in tunneling."""
        return self.get_client(use_zapi=use_zapi_client).get_vserver()

    def set_vserver(self, vserver):
        """Set the vserver to use if tunneling gets enabled."""
        self.zapi_client.set_vserver(vserver)
        self.rest_client.set_vserver(vserver)

    def set_username(self, username):
        """Set the user name for authentication."""
        self.zapi_client.set_username(username)
        self.rest_client.set_username(username)

    def set_password(self, password):
        """Set the password for authentication."""
        self.zapi_client.set_password(password)
        self.rest_client.set_password(password)

    def get_client(self, use_zapi=True):
        """Chooses the client to be used in the request."""
        if use_zapi:
            return self.zapi_client
        return self.rest_client

    def invoke_successfully(self, na_element, api_args=None,
                            enable_tunneling=False, use_zapi=True):
        """Invokes API and checks execution status as success.

        Need to set enable_tunneling to True explicitly to achieve it.
        This helps to use same connection instance to enable or disable
        tunneling. The vserver or vfiler should be set before this call
        otherwise tunneling remains disabled.
        """
        return self.get_client(use_zapi=use_zapi).invoke_successfully(
            na_element, api_args=api_args, enable_tunneling=enable_tunneling)

    def __str__(self):
        return "server: %s" % (self._host)


class NaElement(object):
    """Class wraps basic building block for NetApp API request."""

    def __init__(self, name):
        """Name of the element or etree.Element."""
        if isinstance(name, etree._Element):
            self._element = name
        else:
            self._element = etree.Element(name)

    def get_name(self):
        """Returns the tag name of the element."""
        return self._element.tag

    def set_content(self, text):
        """Set the text string for the element."""
        self._element.text = text

    def get_content(self):
        """Get the text for the element."""
        return self._element.text

    def add_attr(self, name, value):
        """Add the attribute to the element."""
        self._element.set(name, value)

    def add_attrs(self, **attrs):
        """Add multiple attributes to the element."""
        for attr in attrs.keys():
            self._element.set(attr, attrs.get(attr))

    def add_child_elem(self, na_element):
        """Add the child element to the element."""
        if isinstance(na_element, NaElement):
            self._element.append(na_element._element)
            return
        raise ValueError(_("Can only add elements of type NaElement."))

    def get_child_by_name(self, name):
        """Get the child element by the tag name."""
        for child in self._element.iterchildren():
            if child.tag == name or etree.QName(child.tag).localname == name:
                return NaElement(child)
        return None

    def get_child_content(self, name):
        """Get the content of the child."""
        for child in self._element.iterchildren():
            if child.tag == name or etree.QName(child.tag).localname == name:
                return child.text
        return None

    def get_children(self):
        """Get the children for the element."""
        return [NaElement(el) for el in self._element.iterchildren()]

    def has_attr(self, name):
        """Checks whether element has attribute."""
        attributes = self._element.attrib or {}
        return name in attributes.keys()

    def get_attr(self, name):
        """Get the attribute with the given name."""
        attributes = self._element.attrib or {}
        return attributes.get(name)

    def get_attr_names(self):
        """Returns the list of attribute names."""
        attributes = self._element.attrib or {}
        return attributes.keys()

    def add_new_child(self, name, content, convert=False):
        """Add child with tag name and context.

           Convert replaces entity refs to chars.
        """
        child = NaElement(name)
        if convert:
            content = NaElement._convert_entity_refs(content)
        child.set_content(content)
        self.add_child_elem(child)

    @staticmethod
    def _convert_entity_refs(text):
        """Converts entity refs to chars to handle etree auto conversions."""
        text = text.replace("&lt;", "<")
        text = text.replace("&gt;", ">")
        return text

    @staticmethod
    def create_node_with_children(node, **children):
        """Creates and returns named node with children."""
        parent = NaElement(node)
        for child in children.keys():
            parent.add_new_child(child, children.get(child, None))
        return parent

    def add_node_with_children(self, node, **children):
        """Creates named node with children."""
        parent = NaElement.create_node_with_children(node, **children)
        self.add_child_elem(parent)

    def to_string(self, pretty=False, method='xml', encoding='UTF-8'):
        """Prints the element to string."""
        return etree.tostring(self._element, method=method, encoding=encoding,
                              pretty_print=pretty)

    def __getitem__(self, key):
        """Dict getter method for NaElement.

            Returns NaElement list if present,
            text value in case no NaElement node
            children or attribute value if present.
        """

        child = self.get_child_by_name(key)
        if child:
            if child.get_children():
                return child
            else:
                return child.get_content()
        elif self.has_attr(key):
            return self.get_attr(key)
        raise KeyError(_('No element by given name %s.') % (key))

    def __setitem__(self, key, value):
        """Dict setter method for NaElement.

           Accepts dict, list, tuple, str, int, float and long as valid value.
        """
        if key:
            if value:
                if isinstance(value, NaElement):
                    child = NaElement(key)
                    child.add_child_elem(value)
                    self.add_child_elem(child)
                elif isinstance(
                        value,
                        six.string_types + six.integer_types + (float, )):
                    self.add_new_child(key, six.text_type(value))
                elif isinstance(value, (list, tuple, dict)):
                    child = NaElement(key)
                    child.translate_struct(value)
                    self.add_child_elem(child)
                else:
                    raise TypeError(_('Not a valid value for NaElement.'))
            else:
                self.add_child_elem(NaElement(key))
        else:
            raise KeyError(_('NaElement name cannot be null.'))

    def translate_struct(self, data_struct):
        """Convert list, tuple, dict to NaElement and appends.

           Example usage:
           1.
           <root>
               <elem1>vl1</elem1>
               <elem2>vl2</elem2>
               <elem3>vl3</elem3>
           </root>
           The above can be achieved by doing
           root = NaElement('root')
           root.translate_struct({'elem1': 'vl1', 'elem2': 'vl2',
                                  'elem3': 'vl3'})
           2.
           <root>
               <elem1>vl1</elem1>
               <elem2>vl2</elem2>
               <elem1>vl3</elem1>
           </root>
           The above can be achieved by doing
           root = NaElement('root')
           root.translate_struct([{'elem1': 'vl1', 'elem2': 'vl2'},
                                  {'elem1': 'vl3'}])
        """
        if isinstance(data_struct, (list, tuple)):
            for el in data_struct:
                if isinstance(el, (list, tuple, dict)):
                    self.translate_struct(el)
                else:
                    self.add_child_elem(NaElement(el))
        elif isinstance(data_struct, dict):
            for k in data_struct.keys():
                child = NaElement(k)
                if isinstance(data_struct[k], (dict, list, tuple)):
                    child.translate_struct(data_struct[k])
                else:
                    if data_struct[k]:
                        child.set_content(six.text_type(data_struct[k]))
                self.add_child_elem(child)
        else:
            raise ValueError(_('Type cannot be converted into NaElement.'))


class NaApiError(Exception):
    """Base exception class for NetApp API errors."""

    def __init__(self, code='unknown', message='unknown'):
        self.code = code
        self.message = message

    def __str__(self, *args, **kwargs):
        return 'NetApp API failed. Reason - %s:%s' % (self.code, self.message)


def invoke_api(na_server, api_name, api_family='cm', query=None,
               des_result=None, additional_elems=None,
               is_iter=False, records=0, tag=None,
               timeout=0, tunnel=None):
    """Invokes any given API call to a NetApp server.

        :param na_server: na_server instance
        :param api_name: API name string
        :param api_family: cm or 7m
        :param query: API query as dict
        :param des_result: desired result as dict
        :param additional_elems: dict other than query and des_result
        :param is_iter: is iterator API
        :param records: limit for records, 0 for infinite
        :param timeout: timeout seconds
        :param tunnel: tunnel entity, vserver or vfiler name
    """
    record_step = 50
    if not (na_server or isinstance(na_server, NaServer)):
        msg = _("Requires an NaServer instance.")
        raise exception.InvalidInput(reason=msg)
    server = copy.copy(na_server)
    if api_family == 'cm':
        server.set_vserver(tunnel)
    else:
        server.set_vfiler(tunnel)
    if timeout > 0:
        server.set_timeout(timeout)
    iter_records = 0
    cond = True
    while cond:
        na_element = create_api_request(
            api_name, query, des_result, additional_elems,
            is_iter, record_step, tag)
        result = server.invoke_successfully(na_element, True)
        if is_iter:
            if records > 0:
                iter_records = iter_records + record_step
                if iter_records >= records:
                    cond = False
            tag_el = result.get_child_by_name('next-tag')
            tag = tag_el.get_content() if tag_el else None
            if not tag:
                cond = False
        else:
            cond = False
        yield result


def create_api_request(api_name, query=None, des_result=None,
                       additional_elems=None, is_iter=False,
                       record_step=50, tag=None):
    """Creates a NetApp API request.

        :param api_name: API name string
        :param query: API query as dict
        :param des_result: desired result as dict
        :param additional_elems: dict other than query and des_result
        :param is_iter: is iterator API
        :param record_step: records at a time for iter API
        :param tag: next tag for iter API
    """
    api_el = NaElement(api_name)
    if query:
        query_el = NaElement('query')
        query_el.translate_struct(query)
        api_el.add_child_elem(query_el)
    if des_result:
        res_el = NaElement('desired-attributes')
        res_el.translate_struct(des_result)
        api_el.add_child_elem(res_el)
    if additional_elems:
        api_el.translate_struct(additional_elems)
    if is_iter:
        api_el.add_new_child('max-records', six.text_type(record_step))
    if tag:
        api_el.add_new_child('tag', tag, True)
    return api_el



# Copyright (c) 2011-2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Filter support
"""
from oslo_log import log

from manila.scheduler import base_handler

LOG = log.getLogger(__name__)


class BaseFilter(object):
    """Base class for all filter classes."""
    def _filter_one(self, obj, filter_properties):
        """Check if an object passes a filter.

        Return True if it passes the filter, False otherwise.
        Override this in a subclass.
        """
        return True

    def filter_all(self, filter_obj_list, filter_properties):
        """Yield objects that pass the filter.

        Can be overridden in a subclass, if you need to base filtering
        decisions on all objects.  Otherwise, one can just override
        _filter_one() to filter a single object.
        """
        for obj in filter_obj_list:
            if self._filter_one(obj, filter_properties):
                yield obj

    # Set to true in a subclass if a filter only needs to be run once
    # for each request rather than for each instance
    run_filter_once_per_request = False

    def run_filter_for_index(self, index):
        """Check if filter needs to be run for the "index-th" instance.

        Return True if the filter needs to be run for the "index-th"
        instance in a request.  Only need to override this if a filter
        needs anything other than "first only" or "all" behaviour.
        """
        return not (self.run_filter_once_per_request and index > 0)


class BaseFilterHandler(base_handler.BaseHandler):
    """Base class to handle loading filter classes.

    This class should be subclassed where one needs to use filters.
    """

    def get_filtered_objects(self, filter_classes, objs,
                             filter_properties, index=0):
        """Get objects after filter

        :param filter_classes: filters that will be used to filter the
                               objects
        :param objs: objects that will be filtered
        :param filter_properties: client filter properties
        :param index: This value needs to be increased in the caller
                      function of get_filtered_objects when handling
                      each resource.
        """
        list_objs = list(objs)
        LOG.debug("Starting with %d host(s)", len(list_objs))
        for filter_cls in filter_classes:
            cls_name = filter_cls.__name__
            filter_class = filter_cls()

            if filter_class.run_filter_for_index(index):
                objs = filter_class.filter_all(list_objs, filter_properties)
                if objs is None:
                    LOG.debug("Filter %(cls_name)s says to stop filtering",
                              {'cls_name': cls_name})
                    return (None, cls_name)
                list_objs = list(objs)
                msg = ("Filter %(cls_name)s returned %(obj_len)d host(s)"
                       % {'cls_name': cls_name, 'obj_len': len(list_objs)})
                if not list_objs:
                    LOG.info(msg)
                    break
                LOG.debug(msg)
        return (list_objs, cls_name)



# Copyright (c) 2014 eBay Inc.
# Copyright (c) 2015 Rushil Chugh
# Copyright (c) 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common


class ViewBuilder(common.ViewBuilder):
    """Model scheduler-stats API responses as a python dictionary."""

    _collection_name = "scheduler-stats"

    def pool_summary(self, pool):
        """Summary view of a single pool."""
        return {
            'pool': {
                'name': pool.get('name'),
                'host': pool.get('host'),
                'backend': pool.get('backend'),
                'pool': pool.get('pool'),
            }
        }

    def pool_detail(self, pool):
        """Detailed view of a single pool."""
        return {
            'pool': {
                'name': pool.get('name'),
                'host': pool.get('host'),
                'backend': pool.get('backend'),
                'pool': pool.get('pool'),
                'capabilities': pool.get('capabilities'),
            }
        }

    def pools(self, pools, detail=False):
        """View of a list of pools seen by scheduler."""
        view_method = self.pool_detail if detail else self.pool_summary
        return {"pools": [view_method(pool)['pool'] for pool in pools]}
